IFDEF *C2*
FORGET *C2*
}
TAQOZ

pub *C2*	." C2 Debug Interface for Silabs Microcontrollers - 200424.0000 " ;

{
Flash and debug Silabs micros via C2 interface

Peter Jakacki 2012, 2020

References:
AN127 Rev1.4

}



\ Default I/O pins

31	:= C2CK
29	:= C2D

{
// C2 status return codes
#define  INVALID_COMMAND   0x00
#define  COMMAND_FAILED    0x02
#define  COMMAND_OK        0x0D

// C2 interface commands
#define  GET_VERSION       0x01
#define  BLOCK_READ        0x06
#define  BLOCK_WRITE       0x07
#define  PAGE_ERASE        0x08
#define  DEVICE_ERASE      0x03

// C2 Registers
#define  FPDAT             0xB4
#define  FPCTL             0x02
#define  DEVICEID          0x00
#define  REVID             0x01

C2ADD 0x00 RW C2 Address.
The C2ADD register is accessed via the C2 interface. The value written to C2ADD selects the target data register for C2
Data Read and Data Write commands.
0x00: C2DEVID
0x01: C2REVID
0x02: C2FPCTL
0xB4: C2FPDAT

commands
Block Write 0x07 all (flash or EPROM)
Block Read 0x06 all (flash or EPROM)
Page Erase 0x08 flash devices only
Device Erase 0x03 flash devices only
Get Version 0x01 all (flash or EPROM)
Get Derivative 0x02 all (flash or EPROM)
Direct Read 0x09 all (flash or EPROM)
Direct Write 0x0A all (flash or EPROM)
Indirect Read 0x0B all (flash or EPROM)
Indirect Write 0x0C all (flash or EPROM)
}

	*** C2 address constants ***
0	:=  =DEVID
2	:=  =FPCTL
$B4	:=  =FPDAT		( this is $AD on the UB2. Not sure about UB3)
6	:=  =BLKRD

\ Status bits
{
 7	:=  =busy	--- This bit indicates when the EPROM or Flash is busy completing an operation
 6	:=  =eerror	--- EPROM error
 2	:=  =inbusy	--- This bit is set to 1 by the C2 Interface following a write to FPDAT.
			--- It is cleared to 0 when the PI acknowledges the write to FPDAT.
 1	:=  =outready	--- This bit is set to 1 by the PI when output data is available in the FPDAT regisSter

}
4 REG := faddr	--- flash address

pub C2PINS ( clkpin datpin -- \ Assign new pins for the C2 interface )
	AT C2D :=!		\ Rewrite the constant
	AT C2CK :=!
	;

{
sck		long 	spi_ck		' P2 clock pin'
mosi		long 	spi_di		' P2 Data out pin'
miso		long 	spi_do		' P2 Data in pin'
ss		long 	spi_cs		' P2 Slave select pin'
}
pri c2errors	10 REG ;
pub .C2ERR	c2errors W@ ?DUP IF . ELSE ." No " THEN ." Errors " ;

pub C2RESET --- Generate a reset pulse
\	C2CK 2 COG! C2D 3 COG! C2D 4 COG! C2CK 5 COG!
	C2CK PIN
	L 20 us H 5 us
	c2errors W~
	;



{
1.5.2 C2CK Clock Strobes
To generate C2CK clock strobes with a microcontroller-based programmer:
1. Turn the C2CK driver on.
2. Wait at least 40 ns. This helps ensure C2D data setup time.
3. Force C2CK low. Ensure interrupts are disabled at this point.
4. Wait between 80 and 5000 ns.
5. Force C2CK high.
6. Wait at least 120 ns. This helps ensure C2D data valid time
}


pri C2STOP
pri C2START ( -- ) --- Start or stop strobe
	C2D FLOAT L H
	;
{
pri C2STROBE ( -- ) --- Strobe the C2 clock low then high
	L H
	;
}
--- wait for c2d high and flg result - 0 = timeout
pri C2WAIT? ( -- flg )
	C2D FLOAT
	10000 BEGIN L H C2D PIN@ IF EXIT THEN 1- DUP 0= UNTIL
\	  10000 0 DO C2STROBE C2D PIN@ ?LEAVE LOOP
	;
pub C2WAIT	C2WAIT? DROP ;


pri C2WRITE ( data -- )
	C2D 8 FOR L LSBOUT H NEXT
	2DROP
	;

--- Send out the 2-bit C2 instruction
pri C2INS ( instr -- )
	C2D 2 FOR L LSBOUT H NEXT 2DROP
	;

pri C2READ ( -- byte )
	C2D FLOAT \ C2CK PIN

	0 C2D 8 FOR L H MSBIN NEXT DROP 24 >>
\	0 8 FOR 2* C2STROBE C2D PIN@ 1& + NEXT
\	REVB
	;

	*** READ DATA REGISTER ***


pub C2DAT@ ( -- byte ) \ Read the C2 DATA regsiter
---	START	INS	LENGTH	wait	data	STOP
	C2START 0 C2INS 0 C2INS C2WAIT C2READ	C2STOP
	;

	*** WRITE DATA REGISTER ***


pub C2DAT! ( byte -- ) \ Write to the C2 DATA register
---	START	INS	LENGTH	DATA	  wait	STOP
	C2START 1 C2INS 0 C2INS C2WRITE C2WAIT C2STOP
	;

	*** READ ADDRESS/STATUS  REGISTER ***

pub C2STAT? ( -- byte ) --- Read the C2 ADDRESS register with status info
---	START	INS	address	STOP
	C2START 2 C2INS C2READ	C2STOP
	;

	*** WRITE ADDRESS REGISTER ***

pub C2ADR! ( byte -- ) --- Write to the C2 ADDRESS register
---	START	INS	ADDRESS	STOP
	C2START 3 C2INS C2WRITE C2STOP
	;

{
For devices with SFR paging, direct reads through the PI using the Direct Read command are recommended to ensure the SFR Page
is managed properly.
To read an SFR from a device with SFR paging:
1. Perform an Address Write with a value of FPDAT.
2. Write the Direct Read command (0x09) using a Data Write.
3. Poll InBusy until the data is processed by the PI.
4. Poll OutReady until it sets to 1.
5. Perform a Data Read to ensure a return value of 0x0D (no errors).
6. Perform a Data Write with a value of the SFR address.
7. Poll InBusy until the data is processed by the PI.
8. Perform a Data Write with a value of 0x01.
9. Poll InBusy until the data is processed by the PI.
10. Poll OutReady until it sets to 0.
11. Read the SFR value from the device using the Data Read instruction.
}


---  Poll until Out Ready with a timeout
pri C2ORDY ( --  )
	1000 BEGIN 1- C2STAT? 1& OVER 0= OR UNTIL DROP
	;

--- Poll until Out Ready then read DATA regsiter
pub C2DATW@ ( -- byte  )
	C2ORDY C2DAT@
	;

pri c2flg	8 REG ;
pub C2FLG@	C2DATW@ DUP c2flg C! ;

pub C2ERR?	C2FLG@ $0D <> DUP IF c2errors W++ THEN ;
pub C2FLG?	C2FLG@ $0D = DUP NOT IF c2errors W++ THEN ;

pub C2FAIL	." Fail " c2flg C@ .B SPACE ;


pub C2CMD ( cmd -- ) --- This function writes a command to the PI
pub C2DATW! ( byte -- ) ---  Write to DATA register and wait until not busy
	C2DAT!
pri C2IBSY ( -- \ Poll Input Busy )
	1000 BEGIN 1- C2STAT? 2 AND 0= OVER 0= OR UNTIL DROP
	;


{
To read from an SFR on a device that does not have SFR paging:
1. Write the SFR address to the device using the Address Write instruction.
2. Read the SFR value from the device using the Data Read instruction.
}
pub C2REG@ ( adr -- byte )
	C2ADR! C2DAT@
	;
\	$100 0 DO I C2ADR! C2DAT@ I BUFFERS + C! LOOP

{
To write to an SFR on a device that does not have SFR paging:
1. Write the SFR address to the device using the Address Write instruction.
2. Write the SFR value to the device using the Data Write instruction.
For devices with SFR paging, direct writes through the PI using the Direct Write command are recommended to ensure the SFR Page
is managed properly.
}

--- pub C2WriteSFR   \ This function directly writes an SFR and consists of the following steps:
pub C2REG! ( dat reg -- )
	C2ADR! C2DATW!
	;


{
To write to an SFR from a device with SFR paging:
C2ADR!
1. Perform an Address Write with a value of FPDAT.
2. Write the Direct Write command (0x0A) using a Data Write.
3. Poll InBusy until the data is processed by the PI.
4. Poll OutReady it sets to 1.
5. Perform a Data Read to ensure a return value of 0x0D (no errors).

6. Perform a Data Write with a value of the SFR address.
7. Poll InBusy until the data is processed by the PI.
8. Perform a Data Write with a value of 0x01.
9. Poll InBusy until the data is processed by the PI.
10. Perform a Data Write with the new SFR value.
11. Poll InBusy until the data is processed by the PI.
}
--- This function allows writes to SFRs on devices that have SFR Paging and ensures that the SFR page is managed correctly .
pub C2DIR! ( data addr -- )
	=FPDAT C2ADR!
	$0A C2CMD C2IBSY
	C2ERR? IF 2DROP C2FAIL EXIT THEN
 	C2CMD					\ write address
	1 C2CMD
	C2CMD					\ write data
	;




	*** FLASH ***


pub C2HALT
pub C2FLASH --- Place C2 device into Flash programming mode and HALTS CPU
	C2RESET
	5 us
	=FPCTL C2ADR!
	2 C2DATW!
 	4 C2DATW!		\ Halt the core	(AN127Rev1.2)
	1 C2DATW!
	20 ms			\ Charge-pump delay?
	;

	*** C2 DEVICE ERASE ***

{
1. Perform an Address Write with a value of FPDAT.
2. Perform a Data Write with the Device Erase command.
3. Poll on InBusy using Address Read until the bit clears.
4. Poll on OutReady using Address Read until the bit set.
5. Perform a Data Read instruction. A value of 0x0D is okay.
6. Perform a Data Write with a value of 0xDE.
7. Poll on InBusy using Address Read until the bit clears.
8. Perform a Data Write with a value of 0xAD.
9. Poll on InBusy using Address Read until the bit clears.
10. Perform a Data Write with a value of 0xA5.
11. Poll on InBusy using Address Read until the bit clears.
12. Poll on OutReady using Address Read until the bit set.
13. Perform a Data Read instruction. A value of 0x0D is okay.
}
--- erase ALL Flash inc bootloader  (Need C2FLASH set for security before executing)
pub C2ERASE ( -- flg \ Erase all the memory on the C2 device )
\	C2FLASH
---	Device erase command to FPDAT
	=FPDAT C2ADR! 3 C2CMD
	C2ERR? IF FALSE EXIT THEN
	$DE C2DATW!			\ Arming sequence
	$AD C2DATW!
	$A5 C2DATW!			\ Erase commences after this
 	C2DATW@
	;


{
To erase a page of flash: (page typically = 512)
1. Perform an Address Write with a value of FPDAT.
2. Perform a Data Write with the Page Erase command.
3. Poll on InBusy using Address Read until the bit clears.
4. Poll on OutReady using Address Read until the bit set.
5. Perform a Data Read instruction. A value of 0x0D is okay.
6. Perform a Data Write with the page number.
7. Poll on InBusy using Address Read until the bit clears.
8. Poll on OutReady using Address Read until the bit clears.
9. Perform a Data Read instruction. A value of 0x0D is okay.
10. Perform a Data Write with the a value of 0x00.
11. Poll on InBusy using Address Read until the bit clears.
12. Poll on OutReady using Address Read until the bit set.
13. Perform a Data Read instruction. A value of 0x0D is okay.
}

( 512 BYTE PAGE ERASE )

pub C2ERASEP ( addr -- )  \ Erase the page on the C2 device
	C2FLASH
 	=FPDAT C2ADR! 8 C2DATW!		\ Page erase command
	C2ERR? IF DROP FALSE EXIT THEN
	9>> C2DATW!		\ write target page
	C2ERR? IF FALSE EXIT THEN
	0 C2DATW!
	C2ERR? DROP
	;




$4.0000 := C2BUF	--- optional 64KB memory image area

pub C2VERIFY	0 34160 ADO C2BUF I+ C@ I SDC@ <> IF CRLF I .WORD SPACE I SDC@ .BYTE SPACE C2BUF I+ C@ .BYTE THEN KEY ?LEAVE LOOP ;



{	PROGRAM FLASH BLOCK
To program a flash block:
1. Perform an Address Write with a value of FPDAT.
2. Perform a Data Write with the Block Write command.
3. Poll on InBusy using Address Read until the bit clears.
4. Poll on OutReady using Address Read until the bit set.
5. Perform a Data Read instruction. A value of 0x0D is okay.
6. Perform a Data Write with the high byte of the address.
7. Poll on InBusy using Address Read until the bit clears.
8. Perform a Data Write with the low byte of the address.
9. Poll on InBusy using Address Read until the bit clears.
10. Perform a Data Write with the length.
11. Poll on InBusy using Address Read until the bit clears.
12. Perform a Data Write with the data. This will write the data to the flash. Repeat steps 11 and 12 for each byte specified by the
length field
13. Poll on OutReady using Address Read until the bit set.
14. Perform a Data Read instruction. A value of 0x0D is okay.
}


pub C2ProgramBlock ( src addr -- flg  \ Program a 256 byte block )
	CRLF PRINT" C2 Program " OVER .M ."  to " DUP .WORD 3 SPACES
	C2FLASH
--- 1. Perform an Address Write with a value of FPDAT.
 	=FPDAT C2ADR!
--- 2. Perform a Data Write with the Block Write command (7)
	7 C2DATW!						\ BLOCK WRITE COMMAND
--- 3. Poll on InBusy using Address Read until the bit clears.
--- 4. Poll on OutReady using Address Read until the bit set.
--- 5. Perform a Data Read instruction. A value of 0x0D is okay.
	C2FLG?
	  IF
--- 6. Perform a Data Write with the high byte of the address.
	  W>B C2DATW!
--- 7. Poll on InBusy using Address Read until the bit clears.
--- 8. Perform a Data Write with the low byte of the address.
--- 9. Poll on InBusy using Address Read until the bit clears.
	  C2DATW!
--- 10. Perform a Data Write with the length.
--- 11. Poll on InBusy using Address Read until the bit clears.
	  0 C2DATW!		\ block sized address
	  C2FLG?
	    IF
--- 12. Perform a Data Write with the data. This will write the data to the flash. Repeat steps 11 and 12 for each byte specified by the length field
	    256 ADO I DUP 16>> IF C@ ELSE SDC@ THEN C2DATW! LOOP
--- 13. Poll on OutReady using Address Read until the bit set.
	    C2ORDY  TRUE
	    ELSE  FALSE
	    THEN
	  ELSE  FALSE
	  THEN
--- 14. Perform a Data Read instruction. A value of 0x0D is okay.
	;

\ 0FBF0: FF FF FF FF  FF FF FF FF  FF FF FF FF  FF 90 A5 FF     '................' ok

--- Program Flash from memory area but if addr < $10000 then read from file
pub -C2PROGRAM ( addr dst cnt --  )
	C2RESET C2DAT@ $FF = IF CR ." No device " 3DROP EXIT THEN
	3RD 16>> 0= @FILE 0= AND IF ." No File open " 3DROP EXIT THEN
--- Enable flash mode and erase pages
	C2FLASH
	SWAP faddr !
	ADO
--- Avoid invalid memory areas of EFM8UB3
	  faddr @ $9E00 <
	  faddr @ $FA00 $FBFF WITHIN OR
	  IF
	    faddr @ $1FF AND 0= IF faddr @ C2ERASEP THEN
	    I faddr @ C2ProgramBlock $21 + EMIT
	  THEN
	  256 faddr +!
	  256
	+LOOP
	  50 ms
	C2RESET .C2ERR
	;


pub C2ProgramBoot
	" EFM8UB3.ROM" FOPEN$ 0EXIT
	$9A00 $9A00 $400 -C2PROGRAM
	$FA00 $FA00 $200 -C2PROGRAM
	PRINT" EFM Bootloader Programmed "
	;

--- Program the device from a file
pre C2PROGRAM: ( <filename> -- )
	GET$ SPACE
	DUP " BOOT" $=
	IF DROP C2ProgramBoot
	ELSE
	FOPEN$ 0EXIT
	@FILE FSIZE@ SWAP OPEN-SECTOR
	0 0 ROT	-C2PROGRAM
	THEN
	;



{

$02 $B6 C2DIR! $A0 $FF C2DIR!  100 us $02 $EF C2DIR!


WriteDirect(0xB6, 0x02); // FLASH timing adjustment

WriteDirect(0xFF, 0xA0); // VDD monitor init

// Delay 100 μs

WriteDirect(0xEF, 0x02);
}


 *** DEVICE SPECIFC ***
{
pub F99x \ id=25
	C2FLASH
	0 $A7 C2DIR!		\ SFR page 0
 	$8F $B2 C2DIR!		\ OSCICN = enable int osc -
 	0 $A9 C2DIR!		\ CLKSEL = Int Osc - no divide
pri ForceVMON
	C2FLASH
	$A0 $FF C2REG!		\ Vdd supply monitor enable - VDDOK
 	2 $EF C2REG!			\ Enable the VDD Supply Monitor as a reset source.
	;
pub F41x \ id=0C
	C2FLASH
	$87 $B2 C2DIR!		\ Oscillator init
	$10 $B6 C2REG!		\ WriteSFR(0xB6, 0x10) Flash timing
	$10 $c9 C2REG!		\ WriteSFR(0xC9, 0x10) - Voltage regulator init
	ForceVMON
	;
}
{
\ Force a device to erase by setting it's VDD monitor bits etc.
\ Still testing
\ Usage: F99x C2ERASE!!
pub C2ERASE!!
\	C2FLASH F99x  \  2 $EF C2REG! $80 $FF C2REG!
 	\ 0 C2ERASEP DROP
	10 ms
	C2ERASE DROP
 	;
}


 *** C2 DEBUG COMMANDS ***



{
\ ReadBlock type - could be passed to C2ReadBlock if enabled
BYTE c2cmd
pub FLASH 		6 c2cmd C! ;
pub RAM  		$0B c2cmd C! ;
pub CPUREG  	9 c2cmd C! ;
}
\ Read the registers into the buffers using the register block read command
pub C2RB
	C2FLASH
	=FPDAT C2ADR!
	9 C2DATW!
	C2DATW@ DROP
 	$80 C2DATW! $80 C2DATW!
  	BUFFERS $80 ADO C2DATW@ I C! LOOP
	;
pub C2RB?
	C2RB 0 $80
pub BDUMP
	SWAP BUFFERS + SWAP DUMP
	;


\ Write the buffer into the registers using the register block write command
pub C2WB
	C2FLASH
	=FPDAT C2ADR!
	$0A C2DATW!
	C2DATW@ DROP
 	$80 C2DATW! $80 C2DATW!
  	BUFFERS $80 ADO I C@ C2DATW! LOOP
 	;

\ Diagnostic function - check ID of C2 device
pub .C2 ( -- \ Check to see if a C2 device responds and list it's identity codes  )
	C2RESET
	0 C2REG@ ." ID=" .BYTE ." r" 1 C2REG@ .BYTE
--- Diagnostic - dump the C2 SFR registers
pub .C2REGS
	( C2RESET )
	$100 0 DO I C2REG@ I BUFFERS + C! LOOP
 	0 $100 BDUMP
 	;
pub .PORTS
	3 FOR
	CRLF ." P" I PRINT ." :" $80 I+ C2REG@ .BYTE ."  dir:" $A4 I+ C2REG@ .BYTE ."  din:" $F1 I+ C2REG@ .BYTE
	NEXT
	;





{
To read a flash block:
--- 1. Perform an Address Write with a value of FPDAT.
	=FPDAT C2ADR!
--- 2. Perform a Data Write with the Block Read command.
	=BLKRD C2DAT!
--- 3. Poll on InBusy using Address Read until the bit clears.
	C2WAIT
--- 4. Poll on OutReady using Address Read until the bit set.
--- 5. Perform a Data Read instruction. A value of 0x0D is okay.
--- 6. Perform a Data Write with the high byte of the address.
--- 7. Poll on InBusy using Address Read until the bit clears.
--- 8. Perform a Data Write with the low byte of the address.
--- 9. Poll on InBusy using Address Read until the bit clears.
--- 10. Perform a Data Write with the length.
--- 11. Poll on InBusy using Address Read until the bit clears.
REPEAT
12. Poll on OutReady using Address Read until the bit set.
13. Perform a Data Read instruction. This will read the data from the flash.
 Repeat step 12 and 13 for each byte specified by the length field

}

pub C2READBLK ( addr dst -- \ Read a 256 byte block into the dst buffer )
	C2FLASH
	=FPDAT C2ADR!
	6 C2DATW!
	C2ERR? IF ."  REJECT ReadBlock " OVER .W C2RESET EXIT THEN
---	Send Flash Address h l and count = 256
	SWAP W>B C2DATW! C2DATW! 0 C2DATW!
	C2ERR? IF  ."  REJECT ReadBlock parameters " .W CRLF  C2RESET EXIT THEN
	256 ADO C2DATW@ I C! LOOP
	;

pub C2>RAM ( src dst bytes -- )
	ADO DUP I C2READBLK 256 + 256 +LOOP DROP
	;






		*** DEBUG & DIAGNOSTICS ***


	*** FLASH DUMP ***


pub C2FC@ ( addr -- byte )
	DUP >B 0=
	IF
	 ( C2RESET )
	  =FPDAT C2ADR!
	  6 C2DATW!
	  C2FLG? 	\ <> IF ."  REJECT ReadBlock " OVER .W C2RESET EXIT THEN
	  IF
---	Send Flash Address h l and count = 256
	  W>B C2DATW! C2DATW! 0 C2DATW!
	  C2ERR? DROP	\ IF  ."  REJECT ReadBlock parameters " .W CRLF  C2RESET EXIT THEN
	  BUFFERS 256 ADO C2DATW@ I C! LOOP
	  ELSE DROP BUFFERS 256 $AA FILL
	  THEN
	  0
	THEN
	>B BUFFERS + C@
	;


--- DUMP source modifier
pub C2		AT C2FC@ DUP DUP DUMP! C2RESET C2FLASH ;



\ Diagnostic only - displays C2 Flash status
pub C2?   C2STAT? DUP .BYTE SPACE DUP 1& IF ." (Output Ready) " THEN 2 AND IF ." (Input Busy) " THEN ;




{

( ***  INTEL HEX LOADER UTILITIES *** )

\ Will process character by character without any expectations - meaning it can terminate anytime

\ Allocate enough memory for typical 16 bytes/line + headers etc
32 TABLE hexbuf
long hexptr
long dst
long hexadr
byte hexflg
byte hex

\ Write bytes from hexbuf to dest
pri WRITEHEX
	hexbuf C@++ ( ptr cnt )				\ Process 1st byte = count
	SWAP C@++ 8 << ( cnt ptr adrh )		\ next 2 bytes are the destination address
	SWAP C@ +	( cnt addr )
	2DUP + hexadr !					\ Update last unwritten address pointer (external use)
	dst @ + ( cnt dst )				\ add in the offset
	hexbuf 4 + SWAP ROT ( src dst cnt )		\ and write this line to the destination
	CMOVE
	;
pri +HEX ( nibble -- )
	hex C@ 4 << + hex C!
	hexflg C@
	  IF hex C@ hexptr @ C! hexptr ++ hexflg C~
	  ELSE	1 hexflg C+!
	  THEN
	;
pri HEX: ( char -- \ Process Intel Hex input )
	DUP ":" = IF DROP hex C~ hexflg C~ hexbuf hexptr ! EXIT THEN
	DUP "0" "9" WITHIN IF "0" - +HEX EXIT THEN
	DUP "A" "F" WITHIN IF $37 - +HEX EXIT THEN
	$0D = IF hexptr @ hexbuf - 5 > 		\ terminate on an empty line else write a valid line
 	  IF WRITEHEX  ELSE R> DROP THEN  THEN
	;

pub HEXLOAD ( dst -- \ Load an Intel Hex file into the dst area )
	dst !
	BEGIN KEY HEX: AGAIN
	;

{ Quick hex load and program current C2 device
Usage: C2LOAD  Programming device ID=25r01 from 4000 to 0000 for 0400 bytes ok
}
pub C2LOAD
	$4000 HEXLOAD
	dst @ 0 hexadr @ #256 ALIGN
 	."  Programming device " .C2 ."  from " 3RD .WORD SPACE
 	." to 0000 for " DUP .WORD ."  bytes"
	C2PROGRAM
	;
}

{   UB3 ROM

TAQOZ# $9A00 $400 SD DUMP ---
09A00: 02 FA 00 01  00 00 00 12  01 00 02 00  00 00 40 C4     '..............@.'
09A10: 10 CB EA 00  01 00 00 00  01 09 02 22  00 01 01 00     '..........."....'
09A20: 80 32 09 04  00 00 01 03  00 00 00 09  21 11 01 00     '.2..........!...'
09A30: 01 22 1B 00  07 05 81 03  40 00 01 06  00 FF 09 01     '."......@.......'
09A40: A1 01 15 00  26 FF 00 75  08 95 04 09  01 81 02 95     '....&..u........'
09A50: 40 09 01 91  02 C0 E4 F5  08 F5 2A 12  9A 67 BF 24     '@.........*..g.$'
09A60: F5 12 9A 67  8F 2B 22 E5  2A B5 08 05  12 9A F3 80     '...g.+".*.......'
09A70: F6 74 00 25  2A F5 82 E4  34 00 F5 83  E0 FF 05 2A     '.t.%*...4......*'
09A80: 15 2B 22 12  9A 67 8F 25  12 9A 67 8F  26 AE 25 AF     '.+"..g.%..g.&.%.'
09A90: 26 22 8F 0D  75 0B 04 E5  0B 60 08 12  9A D7 12 9A     '&"..u....`......'
09AA0: F3 80 F4 22  75 11 9B 12  9A F3 E5 11  70 F9 F8 F6     '..."u.......p...'
09AB0: 75 EF 12 22  75 A7 20 7D  08 7F 01 71  EA 75 B3 E0     'u.."u. ^...q.u..'
09AC0: 7D 01 7F 07  71 EA 7D 0C  7F 0B 71 EA  7D 8F 7F 0F     '^...q.^...q.^...'
09AD0: 71 EA E4 F5  A7 61 15 7D  01 7F 0E 71  EA 20 00 12     'q....a.^...q. ..'
09AE0: 7F 11 71 D8  EF 20 E0 0A  7F 21 91 6E  7D 01 7F 11     '..q.. ...!.n^...'
09AF0: 71 EA 22 7F  06 71 D8 8F  27 7F 02 71  D8 8F 28 E5     'q."..q..'..q..(.'
09B00: 27 30 E2 02  71 15 E5 27  30 E3 02 05  11 E5 28 30     ''0..q..'0.....(0'
09B10: E0 02 71 22  22 E4 F5 1A  F5 0C D2 00  7D 01 7F 01     '..q"".......^...'
09B20: 61 EA E4 FD  7F 0E 71 EA  7F 11 71 D8  8F 29 E5 29     'a.....q...q..).)'
09B30: 30 E2 04 E4  FD 61 D0 E5  29 30 E4 09  7D 80 7F 11     '0....a..)0..^...'
09B40: 71 EA E4 F5  0C E5 0C 70  4C E5 29 30  E0 47 91 04     'q......pL.)0.G..'
09B50: E5 12 54 7F  FF BF 21 1C  E5 13 24 F7  60 0C 24 08     '..T...!...$.`.$.'
09B60: 70 0D 85 19  0B 75 0C 01  80 0E 75 0C  02 80 09 75     'p....u....u....u'
09B70: 0C 03 80 04  B1 A6 8F 0C  E5 0C B4 03  04 7D 20 80     '.............^ .'
09B80: 0A E5 0C 70  04 7D 48 80  02 7D 40 7F  11 71 EA 7F     '...p.^H..^@..q..'
09B90: 11 71 D8 8F  29 E5 0C 64  01 70 1F E5  29 54 13 70     '.q..)..d.p..)T.p'
09BA0: 19 E5 12 54  7F FF BF 21  06 7F 20 91  6E 80 02 91     '...T...!.. .n...'
09BB0: 2A 7D 0A 7F  11 71 EA E4  F5 0C E5 0C  B4 02 18 E5     '*^...q..........'
09BC0: 29 30 E0 13  7F 16 71 D8  AD 07 7F 20  91 4D 7D 48     ')0....q.... .M^H'
09BD0: 7F 11 71 EA  E4 F5 0C 22  E5 AE 20 E7  FB EF 44 80     '..q....".. ...D.'
09BE0: F5 AE E5 AE  20 E7 FB AF  AF 22 E5 AE  20 E7 FB 8F     '.... ....".. ...'
09BF0: AE 8D AF 22  E5 AE 20 E7  FB 8F AE 22  E5 AE 20 E7     '...".. ....".. .'
09C00: FB AF AF 22  7F 60 71 D8  8F 12 71 FC  8F 13 71 FC     '...".`q...q...q.'
09C10: 8F 15 71 FC  8F 14 71 FC  8F 17 71 FC  8F 16 71 FC     '..q...q...q...q.'
09C20: 8F 19 71 FC  8F 18 E4 FF  61 F4 7F 20  71 F4 E4 FF     '..q.....a.. q...'
09C30: EF C3 95 19  50 16 E5 0A  2F F5 82 E4  35 09 F5 83     '....P.../...5...'
09C40: E4 93 F5 AF  E5 AE 20 E7  FB 0F 80 E4  22 EF 44 C0     '...... .....".D.'
09C50: FF 71 F4 ED  60 14 71 FC  74 00 25 08  F5 82 E4 34     '.q..`.q.t.%....4'
09C60: 00 F5 83 EF  F0 05 08 1D  80 E9 E4 FF  61 F4 71 F4     '............a.q.'
09C70: E4 FF E5 0B  60 11 74 0D  2F F8 E6 F5  AF E5 AE 20     '....`.t./...... '
09C80: E7 FB 15 0B  0F 80 EB 22  75 1F 01 75  09 9A 75 0A     '......."u..u..u.'
09C90: 03 22 75 1F  01 75 09 9A  75 0A 05 22  E5 1A B4 02     '."u..u..u.."....'
09CA0: 14 E5 12 B4  82 0F E5 17  64 81 45 16  70 07 30 00     '........d.E.p.0.'
09CB0: 02 81 88 81  92 E5 17 45  16 70 02 91  92 22 E5 1A     '.......E.p..."..'
09CC0: 64 02 70 2F  E5 12 64 02  70 29 E5 17  64 81 45 16     'd.p/..d.p)..d.E.'
09CD0: 70 21 E5 15  45 14 70 1B  F5 1F 7D 01  7F 0E 71 EA     'p!..E.p...^...q.'
09CE0: E5 13 B4 01  06 C2 00 7D  40 80 04 D2  00 7D 10 7F     '.......^@....^..'
09CF0: 11 71 EA 22  C3 E5 15 94  80 E5 14 94  00 50 12 E4     '.q.".........P..'
09D00: F5 1F AD 15  FF 71 EA E5  15 70 03 F5  1A 22 75 1A     '.....q...p..."u.'
09D10: 01 22 E4 FF  75 1F 01 E5  14 24 FE 60  16 24 E1 60     '."..u....$.`.$.`'
09D20: 1C 14 60 27  24 21 70 2D  75 09 9A 75  0A 07 90 9A     '..`'$!p-u..u....'
09D30: 07 80 13 75  09 9A 75 0A  19 7F 22 80  1B 75 09 9A     '...u..u..."..u..'
09D40: 75 0A 2B 90  9A 2B E4 93  FF 80 0D 75  09 9A 75 0A     'u.+..+.....u..u.'
09D50: 3B 7F 1B 80  03 75 1F 03  EF C3 95 19  50 02 8F 19     ';....u......P...'
09D60: 22 E5 1A B4  02 02 81 88  91 92 22 E5  1A 60 2E C3     '"........."..`..'
09D70: E5 15 94 02  E5 14 94 00  50 23 E4 F5  1F E5 15 60     '........P#.....`'
09D80: 17 75 1A 02  C2 00 7D 01  7F 0E 71 EA  7D 40 7F 11     '.u....^...q.^@..'
09D90: 71 EA 7D 20  7F 12 61 EA  75 1A 01 D2  00 22 E5 1A     'q.^ ..a.u...."..'
09DA0: B4 02 02 91  92 22 75 1F  03 E5 13 B4  0C 00 50 3D     '....."u.......P='
09DB0: 90 9D B6 25  E0 73 A1 CE  A1 D2 A1 ED  A1 D2 A1 ED     '...%.s..........'
09DC0: A1 D6 A1 DA  A1 ED A1 DE  A1 E2 A1 E6  A1 EA 91 9C     '................'
09DD0: 80 1B 91 BE  80 17 91 F4  80 13 B1 12  80 0F B1 61     '...............a'
09DE0: 80 0B B1 6B  80 07 B1 9E  80 03 E4 F5  1F E4 FD 7F     '...k............'
09DF0: 0E 71 EA AF  1F 22 FF FF  FF FF FF FF  FF FF FF FF     '.q..."..........' ok


TAQOZ# $FA00 $200 SD DUMP ---
0FA00: 90 00 00 E4  93 F4 60 18  E5 EF B4 10  08 E8 64 A5     '......`.......d.'
0FA10: 60 0E 02 00  00 54 03 60  F9 78 16 20  A0 F4 D8 FB     '`....T.`.x. ....'
0FA20: 75 81 2B 02  FA A7 12 9A  83 8E 21 8F  22 AD 2B 12     'u.+.......!.".+.'
0FA30: FB 55 50 26  E5 1C B4 02  07 AF 22 AE  21 12 FB 45     '.UP&......".!..E'
0FA40: E5 2B 60 19  12 9A 67 AD  07 AF 22 AE  21 12 FB 4D     '.+`...g...".!..M'
0FA50: 05 22 E5 22  70 EA 05 21  80 E6 75 1B  41 22 12 9A     '."."p..!..u.A"..'
0FA60: 83 8E 21 8F  22 12 9A 83  8E 23 8F 24  43 CE 08 D3     '..!."....#.$C...'
0FA70: E5 22 95 24  E5 21 95 23  50 13 AF 22  AE 21 12 FB     '.".$.!.#P..".!..'
0FA80: 2A 8F DD 05  22 E5 22 70  E6 05 21 80  E2 12 9A 83     '*..."."p..!.....'
0FA90: C0 06 C0 07  12 FB 6D D0  05 D0 04 EF  6D 70 02 EE     '......m.....mp..'
0FAA0: 6C 60 03 75  1B 43 22 12  FB 79 E4 F5  1D F5 1E 12     'l`.u.C"..y......'
0FAB0: 9A 56 12 9A  67 EF 24 D0  F5 1C 75 1B  40 B4 07 00     '.V..g.$...u.@...'
0FAC0: 50 5E 90 FA  C8 25 E0 73  41 D6 41 E6  41 F2 41 F2     'P^...%.sA.A.A.A.'
0FAD0: 41 F7 41 FC  61 16 12 9A  83 EF 70 03  EE 64 36 60     'A.A.a.....p..d6`'
0FAE0: 42 75 1B 42  80 3D 12 9A  67 8F 1D 12  9A 67 8F 1E     'Bu.B.=..g....g..'
0FAF0: 80 31 12 FA  26 80 2C 12  FA 5E 80 27  12 9A 67 AD     '.1..&.,..^.'..g.'
0FB00: 07 7E FB 7F  FE 12 FB 4D  12 9A 67 AD  07 7E FB 7F     '.~.....M..g..~..'
0FB10: FF 12 FB 4D  80 0D 7F 40  12 9A 92 12  9A A4 80 03     '...M...@........'
0FB20: 75 1B 90 AF  1B 12 9A 92  80 85 8F 82  8E 83 E4 93     'u...............'
0FB30: FF 22 8F 82  8E 83 85 1D  B7 85 1E B7  43 8F 01 ED     '."..........C...'
0FB40: F0 53 8F FC  22 43 8F 02  E4 FD 02 FB  32 ED F4 60     '.S.."C......2..`'
0FB50: 03 12 FB 32  22 C3 EE 94  9A 50 10 ED  2F FF E4 3E     '...2"....P../..>'
0FB60: FE D3 EF 94  00 EE 94 9A  50 01 22 C3  22 43 CE 01     '........P."."C..'
0FB70: AF DE EF FE  AD DE ED FF  22 75 97 DE  75 97 AD 43     '........"u..u..C'
0FB80: FF 80 75 EF  02 75 A7 10  43 C1 10 E4  F5 A9 75 A9     '..u..u..C.....u.'
0FB90: 03 75 A7 20  E4 F5 B5 02  9A B4 FF FF  FF FF FF FF     '.u. ............'
0FBA0: FF FF FF FF  FF FF FF FF  FF FF FF FF  FF FF FF FF     '................'
0FBB0: FF FF FF FF  FF FF FF FF  FF FF FF FF  FF FF FF FF     '................'
0FBC0: FF FF FF FF  FF FF FF FF  FF FF FF FF  FF FF FF FF     '................'
0FBD0: FF FF FF FF  FF FF FF FF  FF FF FF FF  FF FF FF FF     '................'
0FBE0: FF FF FF FF  FF FF FF FF  FF FF FF FF  FF FF FF FF     '................'
0FBF0: FF FF FF FF  FF FF FF FF  FF FF FF FF  FF 90 A5 FF     '................' ok


REJECT ReadBlock parameters $9E00 (and above)

}

: C2A	31 29 C2PINS ;
: C2B	33 32 C2PINS ;


END
