IFDEF *LIFE*
_lines org
FORGET *LIFE*
}
TAQOZ

pub *LIFE*        ." Conway's Game of Life for TAQOZ in the P2 ROM V1.0 190226.0000 " ;


{
Adapted from code at:
http://www.forth-ev.de/wiki/doku.php/projects:4e4th:4e4th:start:beispiele

Optimized for cell widths so 32 wide for the Propeller
Modified so that lines print horizontally to optimize the cell width vertically
Widths of up to 512 can be used
Automatically expands horizontally if "expanding" mode is used: 16 WIDE EXPANDING RANDOM LIFE

}


\ SDBUF $800 + := universe
$7E000	:= universe

IFNDEF CELLS pub CELLS		4* ; }


word  _lines

: #lines   _lines W@ ;

8 CELLS		:= bits/cell

pub line  ( n -- a-addr )		CELLS universe + ;

0 TABLE #bits
	$02.01.01.00 ,
	$03.02.02.01 ,
	$02.01.01.00 ,
	$03.02.02.01 ,

long r0
long r1

: ROL3
      2 AND r0 @ OR 1 ROL r0 !
      1 ROL ROT  1 ROL ROT  1 ROL ROT		--- next cell in each line --- 8.4us --- x1 x2 x3 revert to original after 32 ROLs
      ;

pub doline  ( x1 x2 x3 -- x1 x2 x3 x4 )	--- 2.8ms
     0  bits/cell FOR
     r0 !
\ pub alive  ( x1 x2 x3 -- flag ) \ return whether cell at bit1 in line1 is alive in next generation
     OVER 2 AND 0<> r1 !
     ( x1 x2 x3 )
     DUP >N #bits + C@
     ( x1 x2 x3 y )
     3RD >N #bits + C@ +  \ note: cell itself is counted, too. correction below.
     ( x1 x2 x3 y )
     4TH >N #bits + C@ +	--- 6.4us --- 0,1,2,3,4,5,6,7,8,9
     ( x1 x2 x3 y )
     DUP 3 = ( x1 x2 x3 y y=3 ) SWAP 4 = ( x1 x2 x3 y=3 y=4 ) --- 5.8us
     r1 @ AND OR ( x1 x2 x3 y=3|y=4&x2&2 )
     ROL3
     r0 @
   NEXT
   ;



SDBUF $400 +	:= bufgen

--- testing creating variables in code space where we can initialize and also BACKUP
0 TABLE cvars 0 , 0 ,
cvars		:= generation
cvars 2 +	:= ng
cvars 3 +	:= syn
cvars 4 +	:= _exp
cvars 5 +	:= zoom

--- define default characters for alive and dead
'O'	:= live
$20	:= dead
: U	' live 2+ ! ;

pub nextgen  ( -- )
   0 line @   #lines 1- line @  OVER   ( s: line0   x1 x2 )
   0 line
   #lines 4* ADO
      I 4+ @               ( s: line0  x1 x2 x3 )
      doline  I !
      ROT DROP                  ( s: line0  x2 x3 )
   4 +LOOP
   --- special treatment for last line
   ROT doline  #lines 1- line !
   3DROP
   ;


pub gentask
	BEGIN
	  nextgen
	  syn C@ IF BEGIN ng C@ 0= UNTIL THEN				--- wait until the previous buffer has been displayed
	  SDBUF bufgen #lines CELLS CMOVE
	  ng C~~
	 \ zoom C@ #lines <> IF zoom C@ #lines > IF EXPAND ELSE SHRINK THEN THEN
	AGAIN
	;



pub .line  ( bit -- )
   |< #lines 0 DO
     I line @ OVER AND IF live ELSE dead THEN EMIT
   LOOP  DROP
   ;



0 TABLE uclrs	$02.07.01.06 ,	\ cyan | red | white | green |

$20 := wall
yellow := border
pri .WALL	border %PAPER wall EMIT wall EMIT %black PAPER ;
pri .HEAD	border %PAPER wall #lines 4+ EMITS %black %PAPER SPACE CRLF ;

pub SHOW  ( -- )  \ print current life state to console
   OFF %CURSOR %BOLD
   %HOME %blue %PAPER
   ."  Conway's Game of Life - P2 TAQOZ  " #lines PRINT
   ." x32  Generation #" generation W@ PRINT
   ."  every " LAP .LAP 8 SPACES
   CRLF %PLAIN %BOLD %white %PEN %black %PAPER
   .HEAD
   bits/cell FOR
    .WALL
     --- new color
     generation W@ I + 3 AND uclrs + C@ %PEN
     I .line
     %white %PEN .WALL SPACE CRLF
   NEXT
   .HEAD
   PLAIN
   ;


--- commands ---

--- set terminal screen width for matrix
pub WIDE			DUP zoom C!
pub WIDEa			16 MAX 512 MIN _lines W! ;

pub SHRINK			#lines 1- WIDEa ;

pub WIDER			#lines 1+ WIDEa 0 #lines 1- line ! ;

pub EXPAND			WIDER universe DUP 4+ #lines 1- CELLS <CMOVE 0 0 line ! ;


--- modes
pub EXPANDING			_exp C~~ ;

pub SYNCH ( on/off -- )		syn C! ;



--- PATTERN CREATION ---

pub VOID  ( -- )		universe #lines CELLS ERASE ;

pub SEED  ( x1 .. xn n -- )	FOR I line ! NEXT ;

--- some well known patterns:
pub GLIDER  ( -- )		7 1 2  3 SEED ;
pub FPENT  ( -- )		4 $0C 6  3 SEED ;
pub LWSS  ( -- )		$0F $11 1 $12  4 SEED ;
pub DIEHARD			$47 $C0 2  3 SEED ;
pub ACORN			$67 8 $20  3 SEED ;

--- create a random pattern as a seed
--- usage: random life
\ pub BITS? ( n -- bits )		0 BL 0 DO OVER I |< AND IF 1+ THEN LOOP NIP ;

pub RANDOM			#lines FOR RND I line ! NEXT ;

--- Copy a pattern of longs from memory (such as ROM)
--- Usage: $E000 copy life
pub COPY ( src -- )		#lines FOR DUP @ I line ! 4+ NEXT DROP ;


pub MENU
	DUP $61 $7A WITHIN IF $20 - THEN SWITCH
	'R' CASE RANDOM BREAK
	'G' CASE GLIDER BREAK
	'D' CASE DIEHARD BREAK
	'A' CASE ACORN BREAK
	'L' CASE LWSS BREAK
	'F' CASE FPENT BREAK
	'W' CASE WIDE BREAK
	'E' CASE EXPAND BREAK
	'<' CASE SHRINK BREAK
	'>' CASE WIDER BREAK
	$08 CASE VOID BREAK
	$09 CASE %CLS BREAK
	$20 CASE ng C~~ WKEY DROP BREAK
	CASE@ '1' '9' WITHIN IF CASE@ '0' - 4 << zoom W! THEN
	;



pub LIFE
	OFF %CURSOR %CLS OFF SYNCH 64 WIDE
	RANDOM SHOW

	1 NEWCOG 5 ms	' gentask 1 TASK W!
        generation ~
	%HOME LAP LAP
	BEGIN
	  _exp C@ IF
	    0 line @ IF EXPAND THEN
	    #lines 1- line @ IF WIDER THEN
	  THEN
	   syn C@ IF BEGIN ng C@ UNTIL ng C~ THEN
	    SHOW generation W++
	  KEY MENU CASE@ $1B =
	UNTIL
	_exp C~
	CRLF
	syn C~~ ng C~~ 1 COGSTOP
	ON %CURSOR
	;

END
