\   An 8051  disassembler and miscellaneous stuff


\    This file pretends to be an 8051 ( and 8052 )
\    disassembler. It was written in F-PC 2.25 for IBM PC machines.
\
\
\    The disassembler uses a full segment of memory ( 64k )
\    as the image of the 8051 code space.
\
\    This segment should be allocated before the disassembler is used
\    and released before we exit Forth. So the safest way to use
\    way to use this disassembler is to load this file, save the new
\    Forth image, exit Forth ( ignore the warning ) and use the new
\    Forth.  That way the allocation of the 64k will be made by initstuff
\    and the release will be made by bye automatically.

  decimal

      hwords+           \  We activate the hidden words facility
                        \  to hide many of the "junk" words.

                        \  if you are not using F-PC, dont use
                        \  HWORDS+ HEADERLESS HEADERS and BEHEAD
                        \  The code works the same way and you
                        \  will be able to see all the words used.
                        \  ( Many, many, many.... )

      variable     8051-seg         \  pointer to the 8051 "fake memory"

       : get-memory            ( -- )

           defers initstuff
           4096  alloc                     \ requests 64 kbytes of memory
           0 = not
           if
               drop drop ." Not enough memory ! " cr
               ."  Leaving Forth now !!! "       cr
               ."  You cannot use the 8051 Disassembler " cr
               bye
           else
               nip  8051-seg !
           then  ;


       ' get-memory is initstuff       \ insert get-memory in
                                       \ the initialization chain



       : rel-memory           ( -- )     \ release the memory
           8051-seg @ dealloc
           0 = not
           if
                ." Memory release failed !"  cr
                ." You should reboot your PC !"
           then  ;

   \ It is not necessary to reboot the PC, but is safer because your memory
   \ is fragmented and/or corrupted and memory management does not
   \ exist in MS-DOS machines ( in the real meaning of the word ).


     : bye  rel-memory bye ;           ( -- )

         \  release the memory before we exit Forth



\     read-file  ( addr -- )

\     reads a binary file onto the "8051 memory"
\     the starting address is  passed on the stack and the length
\     is the length of the file.
\
\     not giving a file specification Tom's  window file selection tool
\     is activated


        : read-file                 ( addr -- )
              open                    \ open the file
              seqhandle endfile       \ get the lenght ( double number )
              0 = not
              if
                 ." File bigger than 64k bytes "     \ extra precaution
                  close
                  abort
              then

              0.   seek

              dup -rot

              seqhandle  8051-seg @  exhread
              = not
                if
                  ." Error reading file "
                then
              close  ;

\     write-file    ( addr len -- )

\     writes a binary file from  the "8051 memory",
\     the  starting address and the length are on the stack.
\



handle outhcb                   \ output file we are writing to.


: write-file    ( addr len  --- )
                dup -rot
                70 rmargin !                             \ set right margin
                outhcb !hcb outhcb 1+ c@ ascii . =       \ of null filename
                if      cr ." Name of file to write? "  \ prompt for filenam
                        query outhcb !hcb                \ stuff in outhcb
                then
                outhcb hcreate abort" Create ERROR!"    \ make output file
                0.0 outhcb movepointer                  \ reset file pointer

                outhcb 8051-seg @ exhwrite
                <> if
                      ."    Write error !! "  abort
                   then
                outhcb hclose drop ;                    \ close out






       \    how to acess easily the "8051 memory"


        :   x! 8051-seg @ swap  !l ;      ( n addr --- )
        :   x@ 8051-seg @ swap  @l ;      ( addr -- n  )

        :   xc! 8051-seg @ swap  c!l ;    ( c addr --- )
        :   xc@ 8051-seg @ swap  c@l ;    ( addr -- c  )


 \ note:  Here starts the portable (?) code !
 \
 \
 \
 \         Porting this code to other Forths
 \
 \ If you dont have a MS-DOS machine ( sadly the standand )
 \ or you dont have F-PC ( happily the standard )
 \ you must have a memory area reserved for the fake 8051 memory.
 \ The ideal is 64 kbytes , but a 8051 program should not have
 \ 64 kbytes !!!  So, in pratice 8 kbytes will be enough for the
 \ majority of the users.
 \
 \  You should grab a memory area ( somewhere ) and make the
 \  the following words by your own:
 \
 \       read-file :  Reads a binary file ( easy )
 \                    to that memory area at an "address" given
 \                    by the variable load-offset.
 \                     ( reads the entire file )
 \
 \      write-file :  Writes a binary file from the fake 8051
 \                    memory, starting at addr and with length len.
 \                    Prompts for the filename to be used.
 \
 \   x! x@ xc! xc@ :  Word and byte store and fetch from your
 \                    memory area, using "fake" 8051 addresses.
 \
 \   So if you only have 8 kbytes you can "put them" in any area of
 \   the 8051's 64 kbytes by incorporating an offset in these words.
 \
 \    All the remaining code depends on these words for simulating
 \    the 8051's memory





  :  ext-field                      ( w n1 n2 -- w1 )
      >r                            \ to extract from a word
      >>                        \ a field that starts at bit n1
      -1 r> <<                  \ and has n2 bits of width
      not and ;


  : .udec                           ( u -- )
      base @ >r                     \ to print a number
      decimal                       \ in decimal
      (u.) type                     \ without disturbing the current base
      r> base !    ;                \  note: does not print any space !!

  : .dec                           ( n -- )
      base @ >r                     \ to print a number
      decimal                       \ in decimal
      (.) type                      \ without disturbing the current base
      r> base !    ;                \  note: does not print any space !!


 : .h	.B 'h' EMIT ;



 : .4#h                         ( n -- )
            dup
            $9F00  u> if
                         ascii 0 emit
                     then
            s>d  <# # # # # #> type

            ascii h emit  ;

      \ prints a word using four digits and with a trailing h
      \ if the most significative digit is a letter A..F prints
      \ a preceeding 0


 defer  .2#hex
 defer  .4#hex

 defer  .byte
 defer  .word




 : bytes-dec    [']  .udec   is  .byte  ;   ( -- )
 : bytes-hex    [']  .2#hex  is  .byte  ;    ( -- )

 \ we can print a byte in decimal or in hexa

   bytes-dec      \ by default


 : words-dec    [']  .udec    is  .word  ;   ( -- )
 : words-hex    [']  .4#hex   is  .word  ;    ( -- )

 \ we can print a word in decimal or in hexa

   words-hex      \ by default

 : hex-$                     ( -- )
                [']  .B  is  .2#hex
                [']  .W  is  .4#hex   ;

 : hex-h                     ( -- )
                [']  .h  is  .2#hex
                [']  .4#h  is  .4#hex   ;

 \ to print in a number in hexa we can use a preceeding $ --> $A0
 \ ( hex-$ )  or a trailing h   --> 0A0h   ( hex-h )
 \
   hex-h     \ by default


 :  goto-col              ( col# -- )      \ goto a specified column in
      #out @ - dup                         \ the same line
      0< if  drop                          \ note: the variable #out
         else spaces                       \ contains the current column
         then   ;                          \ position

 \ yes: this function only advances the cursor
 \      to a specified column, it will not go back to a specified
 \      column

  \ some chars very used  on the 8051's instruction set

   : accumulator        'A' emit ;     ( -- )

   : cardinal           '#' emit ;     ( -- )

   : ampersand          '@' EMIT ;     ( -- )

   : carry              'C' EMIT ;     ( -- )

   : comma              ',' EMIT ;     ( -- )


      long     inst-pointer     \  "address"  of the current instruction


      byte     #bytes           \   number of bytes of each instruction


     defer name      ( -- )        \ word that will print
                                   \ the name of each instruction
      ' noop is name

     defer operands  ( -- )        \ word that will print the
                                   \ operands of each instruction
      ' noop is operands





   : 1byte                    ( -- b )
        inst-pointer @
        xc@ ;
 \ get the first byte of the instruction

   : 2byte                    ( -- b )
        inst-pointer @  1 +
        xc@ ;
 \ get the second byte of the instruction

   : 3byte                    ( -- b )
        inst-pointer @  2 +
        xc@ ;

 \ get the third byte of the instruction



   headerless     \ this is to hide the majority the words
                  \ used, not necessary, but a good thing if you have it


  \ now some words to help on the printing of the operands


   : a_addr                     ( -- )
        1byte                  \ print the target of an absolute jmp
        3 <<               \ or call instruction
        $ff00  and
        2byte or
        inst-pointer @         \ weird adressing mode
        $f800 and              \ heritage from the 8048
        or
        .word   ;



   : .internal-byte             ( b -- )
                          \ some internal bytes have
                          \ name with a meaning
       case
        $e0     of      ." ACC"         endof
        $f0     of      ." B"           endof
        $d0     of      ." PSW"         endof
        $81     of      ." SP"          endof
        $82     of      ." DPL"         endof
        $83     of      ." DPH"         endof
        $80     of      ." P0"          endof
        $90     of      ." P1"          endof
        $a0     of      ." P2"          endof
        $b8     of      ." IP"          endof
        $a8     of      ." IE"          endof
        $89     of      ." TMOD"        endof
        $8c     of      ." TH0"         endof
        $8a     of      ." TL0"         endof
        $8d     of      ." TH1"         endof
        $8b     of      ." TL1"         endof
        $98     of      ." SCON"        endof
        $99     of      ." SBUF"        endof
        $87     of      ." PCON"        endof

        .byte

       endcase ;


   : .internal-bit        ( b -- )

                            \ some internal bits have a name with
         case               \ a meaning
                \  P0
        $80     of      ." P0.0"        endof
        $81     of      ." P0.1"        endof
        $82     of      ." P0.2"        endof
        $83     of      ." P0.3"        endof
        $84     of      ." P0.4"        endof
        $85     of      ." P0.5"        endof
        $86     of      ." P0.6"        endof
        $87     of      ." P0.7"        endof
                \  TCON
	    $88     of      ." IT0"         endof
        $89     of      ." IE0"         endof
        $8A     of      ." IT1"         endof
        $8B     of      ." IE1"         endof
        $8C     of      ." TR0"         endof
        $8D     of      ." TF0"         endof
        $8E     of      ." TR1"         endof
        $8F     of      ." TF1"         endof
                \  P1
        $90     of      ." P1.0"        endof
        $91     of      ." P1.1"        endof
        $92     of      ." P1.2"        endof
        $93     of      ." P1.3"        endof
        $94     of      ." P1.4"        endof
        $95     of      ." P1.5"        endof
        $96     of      ." P1.6"        endof
        $97     of      ." P1.7"        endof

               \ SCON
        $98     of      ." RI"          endof
        $99     of      ." TI"          endof
        $9A     of      ." RB8"         endof
        $9B     of      ." TB8"         endof
        $9C     of      ." REN"         endof
        $9D     of      ." SM2"         endof
        $9E     of      ." SM1"         endof
        $9F     of      ." SM0"         endof


                \  P2
        $A0     of      ." P2.0"        endof
        $A1     of      ." P2.1"        endof
        $A2     of      ." P2.2"        endof
        $A3     of      ." P2.3"        endof
        $A4     of      ." P2.4"        endof
        $A5     of      ." P2.5"        endof
        $A6     of      ." P2.6"        endof
        $A7     of      ." P2.7"        endof

                \ IE
        $A8     of      ." EX0"         endof
        $A9     of      ." ET0"         endof
        $AA     of      ." EX1"         endof
        $AB     of      ." ET1"         endof
        $AC     of      ." ES"          endof
        $AD     of      ." ET2"         endof
        $AE     of      ." IE.6"        endof
        $AF     of      ." EA"          endof

                \  P3
        $B0     of      ." P3.0"        endof
        $B1     of      ." P3.1"        endof
        $B2     of      ." P3.2"        endof
        $B3     of      ." P3.3"        endof
        $B4     of      ." P3.4"        endof
        $B5     of      ." P3.5"        endof
        $B6     of      ." P3.6"        endof
        $B7     of      ." P3.7"        endof

                \  IP
        $B8     of      ." PX0"         endof
        $B9     of      ." PT0"         endof
        $BA     of      ." PX1"         endof
        $BB     of      ." PT1"         endof
        $BC     of      ." PS"          endof
        $BD     of      ." PT2"         endof    \ 8052 only
        $BE     of      ." IP.6"        endof
        $BF     of      ." IP.7"        endof

                \ T2CON         8052 only
        $C8     of      ." CP/-RL2"     endof
        $C9     of      ." C/-T2"       endof
        $CA     of      ." TR2"         endof
        $CB     of      ." EXEN2"       endof
        $CC     of      ." TCLK"        endof
        $CD     of      ." RCLK"        endof
        $CE     of      ." EXF2"        endof
        $CF     of      ." TF2"         endof

                \  PSW
        $D0     of      ." P"           endof
        $D1     of      ." PSW.1"       endof
        $D2     of      ." OV"          endof
        $D3     of      ." RS0"         endof
        $D4     of      ." RS1"         endof
        $D5     of      ." F0"          endof
        $D6     of      ." AC"          endof
        $D7     of      ." CY"          endof

               \  ACC
        $E0     of      ." ACC.0"       endof
        $E1     of      ." ACC.1"       endof
        $E2     of      ." ACC.2"       endof
        $E3     of      ." ACC.3"       endof
        $E4     of      ." ACC.4"       endof
        $E5     of      ." ACC.5"       endof
        $E6     of      ." ACC.6"       endof
        $E7     of      ." ACC.7"       endof

               \  B register

        $F0     of      ." B.0"       endof
        $F1     of      ." B.1"       endof
        $F2     of      ." B.2"       endof
        $F3     of      ." B.3"       endof
        $F4     of      ." B.4"       endof
        $F5     of      ." B.5"       endof
        $F6     of      ." B.6"       endof
        $F7     of      ." B.7"       endof

          .byte

       endcase ;


   : rn                         ( -- )
        ascii R emit
        1byte
        7 and
        .udec ;

   : @ri                        ( -- )
        ampersand
        ascii R emit
        1byte
        1 and
        .udec  ;

    :  a,rn                     ( -- )

       accumulator
       comma
       rn  ;

    : a,@ri                       ( -- )

       accumulator
       comma
       @ri    ;

    : a,#byte                   ( -- )

       accumulator
       comma
       cardinal
       2byte .byte   ;

    : a,byte                    ( -- )
        accumulator
        comma
        2byte .internal-byte   ;


    : c,/bit                   ( -- )
         ascii C emit
         comma
         ascii / emit
         2byte .internal-bit ;

    : c,bit                     ( -- )
         ascii C emit
         comma
         2byte .internal-bit ;

    : byte,a
         2byte .internal-byte
         comma
         accumulator ;

    : byte,#byte                 ( -- )
         2byte .internal-byte
         comma
         cardinal
         3byte .byte ;

    : pc_relative   ( n -- )
         dup
         127 >
         if
           $ff00 or      \ sign extend
         then
         dup
         .dec
         3 spaces
         inst-pointer @  +
         #bytes  @ +
         43 goto-col
         base @ >r
         hex  .4#hex
         r> base ! ;


    : a,#byte,byte             ( -- )
         accumulator
         comma
         cardinal
         2byte .byte
         comma
         3byte pc_relative ;

    : a,byte,byte              ( -- )
         accumulator
         comma
         2byte .internal-byte
         comma
         3byte pc_relative ;

    : rn,#byte,byte            ( -- )
         rn
         comma
         cardinal
         2byte .byte
         comma
         3byte pc_relative ;

    : @ri,#byte,byte           ( -- )
         @ri
         comma
         cardinal
         2byte .byte
         comma
         3byte pc_relative ;

    : byte                     ( -- )
         2byte .internal-byte ;

    : #byte                     ( -- )
         2byte .byte ;

    : bit                     ( -- )
         2byte .internal-bit ;


    : rn,pcrel                ( -- )
         rn
         comma
         2byte
         pc_relative ;

    : bit,pcrel              ( -- )
         2byte
         .internal-bit
         comma
         3byte
         pc_relative ;

    : byte,pcrel              ( -- )
         2byte
         .internal-byte
         comma
         3byte
         pc_relative ;

    : pcrel                   ( -- )
         2byte
         pc_relative ;


     : addr                    ( -- )
          2byte 8 << 3byte or
            base @ >r hex
            .4#hex
            r> base ! ;


     : dptr,#addr              ( -- )
         ." DPTR,#"
          addr ;

     : rn,a                     ( -- )
         rn
         comma
         accumulator ;

     : rn,#byte                 ( -- )
         rn
         comma
         cardinal
         #byte ;

     : rn,byte                  ( -- )
         rn
         comma
         byte ;

     : @ri,a                    ( -- )
         @ri
         comma
         accumulator ;

     : @ri,byte                 ( -- )
         @ri
         comma
         byte ;

     : @ri,#byte                ( -- )
         @ri
         comma
         cardinal
         #byte ;

     : bit,c                    ( -- )
         2byte .internal-bit
         comma
         ascii C emit ;

     : byte,rn                   ( -- )
         byte
         comma
         rn ;

     : byte,@ri                    ( -- )
         byte
         comma
         @ri   ;

     : byte,byte                 ( -- )
         3byte
         .internal-byte
         comma
         2byte
         .internal-byte
          ;

     : a,@a+dptr
         ." A,@A+DPTR"  ;

     : a,@a+pc
         ." A,@A+PC"    ;

     : a,@dptr
         ." A,@DPTR"    ;

     : @dptr,a
         ." @DPTR,A"    ;

     : @a+dptr
         ." @A+DPTR"    ;

     : $dptr
            ." DPTR"    ;


     \   instruction  names


     : $acall     ." ACALL"             ;       ( -- )

     : $add       ." ADD"               ;       ( -- )

     : $addc      ." ADDC"              ;       ( -- )

     : $ajmp      ." AJMP"              ;       ( -- )

     : $anl       ." ANL"               ;       ( -- )

     : $cjne      ." CJNE"              ;       ( -- )

     : $clr       ." CLR"               ;       ( -- )

     : $cpl       ." CPL"               ;       ( -- )

     : $daa       ." DA    A"           ;       ( -- )

     : $dec       ." DEC"               ;       ( -- )

     : $divab     ." DIV   AB"          ;       ( -- )

     : $djnz      ." DJNZ"              ;       ( -- )

     : $inc       ." INC"               ;       ( -- )

     : $jb        ." JB"                ;       ( -- )

     : $jbc       ." JBC"               ;       ( -- )

     : $jc        ." JC"                ;       ( -- )

     : $jmp       ." JMP"               ;       ( -- )

     : $jnb       ." JNB"               ;       ( -- )

     : $jnc       ." JNC"               ;       ( -- )

     : $jnz       ." JNZ"               ;       ( -- )

     : $jz        ." JZ"                ;       ( -- )

     : $lcall     ." LCALL"             ;       ( -- )

     : $ljmp      ." LJMP"              ;       ( -- )

     : $mov       ." MOV"               ;       ( -- )

     : $movc      ." MOVC"              ;       ( -- )

     : $movx      ." MOVX"              ;       ( -- )

     : $mulab     ." MUL   AB"          ;       ( -- )

     : $nop       ." NOP"               ;       ( -- )

     : $orl       ." ORL"               ;       ( -- )

     : $pop       ." POP"               ;       ( -- )

     : $push      ." PUSH"              ;       ( -- )

     : $ret       ." RET"               ;       ( -- )

     : $reti      ." RETI"              ;       ( -- )

     : $rla       ." RL    A"           ;       ( -- )

     : $rlca      ." RLC   A"           ;       ( -- )

     : $rra       ." RR    A"           ;       ( -- )

     : $rrca      ." RRC   A"           ;       ( -- )

     : $setb      ." SETB"              ;       ( -- )

     : $sjmp      ." SJMP"              ;       ( -- )

     : $subb      ." SUBB"              ;       ( -- )

     : $swapa     ." SWAP  A"           ;       ( -- )

     : $xch       ." XCH"               ;       ( -- )

     : $xchd      ." XCHD"              ;       ( -- )

     : $xrl       ." XRL"               ;       ( -- )

     : $illop     ." ILLOP"             ;       ( -- )


pri instr
	R>
	DUP @ #bytes C!
	DUP 2_ @ is operands
	4+ @ is name
	;

pri acall_addr	instr 2 a_addr $acall
pri add_a,rn	instr 1 a,rn $add



     : instruction   create   , , ,      ( 'name  'operands  #bytes -- )

                     does>               ( -- )
                     dup
                     @ #bytes !
                     dup
                     2 + @ is operands
                     4 + @ is name ;

    \ this is the compiling word that generates the "instruccions"
    \ each "instruction" must place the correct number on the
    \ #bytes variable, and set name and operands to the apropriate words


     ' $acall ' a_addr          2 instruction acall_addr

     ' $add   ' a,rn            1 instruction add_a,rn

     ' $add   ' a,@ri           1 instruction add_a,@ri

     ' $add   ' a,#byte         2 instruction add_a,#byte

     ' $add   ' a,byte          2 instruction add_a,byte

     ' $addc  ' a,rn            1 instruction addc_a,rn

     ' $addc  ' a,@ri           1 instruction addc_a,@ri

     ' $addc  ' a,#byte         2 instruction addc_a,#byte

     ' $addc  ' a,byte          2 instruction addc_a,byte

     ' $ajmp  ' a_addr          2 instruction ajmp_addr

     ' $anl   ' a,rn            1 instruction anl_a,rn

     ' $anl   ' a,@ri           1 instruction anl_a,@ri

     ' $anl   ' a,#byte         2 instruction anl_a,#byte

     ' $anl   ' a,byte          2 instruction anl_a,byte

     ' $anl   ' c,/bit          2 instruction anl_c,/bit

     ' $anl   ' c,bit           2 instruction anl_c,bit

     ' $anl   ' byte,a          2 instruction anl_byte,a

     ' $anl   ' byte,#byte      3 instruction anl_byte,#byte

     ' $cjne  ' a,#byte,byte    3 instruction cjne_a,#byte,byte

     ' $cjne  ' a,byte,byte     3 instruction cjne_a,byte,byte

     ' $cjne  ' rn,#byte,byte   3 instruction cjne_rn,#byte,byte

     ' $cjne  ' @ri,#byte,byte  3 instruction cjne_@ri,#byte,byte

     ' $clr   ' accumulator     1 instruction clr_a

     ' $clr   ' carry           1 instruction clr_c

     ' $clr   ' bit             2 instruction clr_bit

     ' $cpl   ' accumulator     1 instruction cpl_a

     ' $cpl   ' carry           1 instruction cpl_c

     ' $cpl   ' bit             2 instruction cpl_bit

     ' $daa   ' noop            1 instruction daa

     ' $dec   ' accumulator     1 instruction dec_a

     ' $dec   ' rn              1 instruction dec_rn

     ' $dec   ' @ri             1 instruction dec_@ri

     ' $dec   ' byte            2 instruction dec_byte

     ' $divab ' noop            1 instruction divab

     ' $djnz  ' rn,pcrel        2 instruction djnz_rn,pcrel

     ' $djnz  ' byte,pcrel      3 instruction djnz_byte,pcrel

     ' $inc   ' accumulator     1 instruction inc_a

     ' $inc   ' rn              1 instruction inc_rn

     ' $inc   ' @ri             1 instruction inc_@ri

     ' $inc   ' $dptr           1 instruction inc_dptr

     ' $inc   ' byte            2 instruction inc_byte

     ' $jb    ' bit,pcrel       3 instruction jb_bit,pcrel

     ' $jbc   ' bit,pcrel       3 instruction jbc_bit,pcrel

     ' $jc    ' pcrel           2 instruction jc_pcrel

     ' $jmp   ' @a+dptr         1 instruction jmp_@a+dptr

     ' $jnb   ' bit,pcrel       3 instruction jnb_bit,pcrel

     ' $jnc   ' pcrel           2 instruction jnc_pcrel

     ' $jnz   ' pcrel           2 instruction jnz_pcrel

     ' $jz    ' pcrel           2 instruction jz_pcrel

     ' $lcall ' addr            3 instruction lcall_addr

     ' $ljmp  ' addr            3 instruction ljmp_addr

     ' $mov   ' a,rn            1 instruction mov_a,rn

     ' $mov   ' a,@ri           1 instruction mov_a,@ri

     ' $mov   ' a,#byte         2 instruction mov_a,#byte

     ' $mov   ' a,byte          2 instruction mov_a,byte

     ' $mov   ' c,bit           2 instruction mov_c,bit

     ' $mov   ' dptr,#addr      3 instruction mov_dptr,#addr

     ' $mov   ' rn,a            1 instruction mov_rn,a

     ' $mov   ' rn,#byte        2 instruction mov_rn,#byte

     ' $mov   ' rn,byte         2 instruction mov_rn,byte

     ' $mov   ' @ri,a           1 instruction mov_@ri,a

     ' $mov   ' @ri,#byte       2 instruction mov_@ri,#byte

     ' $mov   ' @ri,byte        2 instruction mov_@ri,byte

     ' $mov   ' byte,a          2 instruction mov_byte,a

     ' $mov   ' bit,c           2 instruction mov_bit,c

     ' $mov   ' byte,rn         2 instruction mov_byte,rn

     ' $mov   ' byte,@ri        2 instruction mov_byte,@ri

     ' $mov   ' byte,#byte      3 instruction mov_byte,#byte

     ' $mov   ' byte,byte       3 instruction mov_byte,byte

     ' $movc  ' a,@a+dptr       1 instruction movc_a,@a+dptr

     ' $movc  ' a,@a+pc         1 instruction movc_a,@a+pc

     ' $movx  ' a,@ri           1 instruction movx_a,@ri

     ' $movx  ' a,@dptr         1 instruction movx_a,@dptr

     ' $movx  ' @ri,a           1 instruction movx_@ri,a

     ' $movx  ' @dptr,a         1 instruction movx_@dptr,a

     ' $mulab ' noop            1 instruction mul_ab

     ' $nop   ' noop            1 instruction nop

     ' $orl   ' a,rn            1 instruction orl_a,rn

     ' $orl   ' a,@ri           1 instruction orl_a,@ri

     ' $orl   ' a,#byte         2 instruction orl_a,#byte

     ' $orl   ' a,byte          2 instruction orl_a,byte

     ' $orl   ' c,/bit          2 instruction orl_c,/bit

     ' $orl   ' c,bit           2 instruction orl_c,bit

     ' $orl   ' byte,a          2 instruction orl_byte,a

     ' $orl   ' byte,#byte      3 instruction orl_byte,#byte

     ' $pop   ' byte            2 instruction pop_byte

     ' $push  ' byte            2 instruction push_byte

     ' $ret  ' noop             1 instruction ret

     ' $reti  ' noop            1 instruction reti

     ' $rla   ' noop            1 instruction rla

     ' $rlca  ' noop            1 instruction rlca

     ' $rra   ' noop            1 instruction rra

     ' $rrca  ' noop            1 instruction rrca

     ' $setb  ' carry           1 instruction setb_c

     ' $setb  ' bit             2 instruction setb_bit

     ' $sjmp ' pcrel            2 instruction sjmp_pcrel

     ' $subb ' a,rn             1 instruction subb_a,rn

     ' $subb ' a,@ri            1 instruction subb_a,@ri

     ' $subb ' a,#byte          2 instruction subb_a,#byte

     ' $subb ' a,byte           2 instruction subb_a,byte

     ' $swapa ' noop            1 instruction swapa

     ' $xch   ' a,rn            1 instruction xch_a,rn

     ' $xch   ' a,@ri           1 instruction xch_a,@ri

     ' $xch   ' a,byte          2 instruction xch_a,byte

     ' $xchd  ' a,@ri           1 instruction xchd_a,@ri

     ' $xrl   ' a,rn            1 instruction xrl_a,rn

     ' $xrl   ' a,@ri           1 instruction xrl_a,@ri

     ' $xrl   ' a,#byte         2 instruction xrl_a,#byte

     ' $xrl   ' a,byte          2 instruction xrl_a,byte

     ' $xrl   ' byte,a          2 instruction xrl_byte,a

     ' $xrl   ' byte,#byte      3 instruction xrl_byte,#byte

     ' $illop ' noop            1 instruction illop





 \  instruction  opcode table
 \
        create  opcode-table

        hex

        \ mask     opcode     token

        01F     c,      011     c,      000     c,      \ acall_addr

        0f8     c,      028     c,      001     c,      \ add_a,rn

        0fe     c,      026     c,      002     c,      \ add_a,@ri

        0ff     c,      024     c,      003     c,      \ add_a,#byte

        0ff     c,      025     c,      004     c,      \ add_a,byte

        0f8     c,      038     c,      005     c,      \ addc_a,rn

        0fe     c,      036     c,      006     c,      \ addc_a,@ri

        0ff     c,      034     c,      007     c,      \ addc_a,#byte

        0ff     c,      035     c,      008     c,      \ addc_a,byte

        01f     c,      001     c,      009     c,      \ ajmp_addr

        0f8     c,      058     c,      00a     c,      \ anl_a,rn

        0fe     c,      056     c,      00b     c,      \ anl_a,@ri

        0ff     c,      054     c,      00c     c,      \ anl_a,#byte

        0ff     c,      055     c,      00d     c,      \ anl_a,byte

        0ff     c,      0b0     c,      00e     c,      \ anl_c,/bit

        0ff     c,      082     c,      00f     c,      \ anl_c,bit

        0ff     c,      052     c,      010     c,      \ anl_byte,a

        0ff     c,      053     c,      011     c,      \ anl_byte,#byte

        0ff     c,      0b4     c,      012     c,      \ cjne_a,#byte,byte

        0ff     c,      0b5     c,      013     c,      \ cjne_a,byte,byte

        0f8     c,      0b8     c,      014     c,      \ cjne_rn,#byte,byte

        0fe     c,      0b6     c,      015     c,      \ cjne_@ri,#byte,byte

        0ff     c,      0e4     c,      016     c,      \ clr_a

        0ff     c,      0c3     c,      017     c,      \ clr_c

        0ff     c,      0c2     c,      018     c,      \ clr_bit

        0ff     c,      0f4     c,      019     c,      \ cpl_a

        0ff     c,      0b3     c,      01a     c,      \ cpl_c

        0ff     c,      0b2     c,      01b     c,      \ cpl_bit

        0ff     c,      0d4     c,      01c     c,      \ daa

        0ff     c,      014     c,      01d     c,      \ dec_a

        0f8     c,      018     c,      01e     c,      \ dec_rn

        0fe     c,      016     c,      01f     c,      \ dec_@ri

        0ff     c,      015     c,      020     c,      \ dec_byte

        0ff     c,      084     c,      021     c,      \ divab

        0f8     c,      0d8     c,      022     c,      \ djnz_rn,pcrel

        0ff     c,      0d5     c,      023     c,      \ djnz_byte,pcrel

        0ff     c,      004     c,      024     c,      \ inc_a

        0f8     c,      008     c,      025     c,      \ inc_rn

        0fe     c,      006     c,      026     c,      \ inc_@ri

        0ff     c,      0a3     c,      027     c,      \ inc_dptr

        0ff     c,      005     c,      028     c,      \ inc_byte

        0ff     c,      020     c,      029     c,      \ jb_bit,pcrel

        0ff     c,      010     c,      02a     c,      \ jbc_bit,pcrel

        0ff     c,      040     c,      02b     c,      \ jc_pcrel

        0ff     c,      073     c,      02c     c,      \ jmp_@a+dptr

        0ff     c,      030     c,      02d     c,      \ jnb_bit,pcrel

        0ff     c,      050     c,      02e     c,      \ jnc_pcrel

        0ff     c,      070     c,      02f     c,      \ jnz_pcrel

        0ff     c,      060     c,      030     c,      \ jz_pcrel

        0ff     c,      012     c,      031     c,      \ lcall_addr

        0ff     c,      002     c,      032     c,      \ ljmp_addr

        0f8     c,      0e8     c,      033     c,      \ mov_a,rn

        0fe     c,      0e6     c,      034     c,      \ mov_a,@ri

        0ff     c,      074     c,      035     c,      \ mov_a,#byte

        0ff     c,      0e5     c,      036     c,      \ mov_a,byte

        0ff     c,      0a2     c,      037     c,      \ mov_c,bit

        0ff     c,      090     c,      038     c,      \ mov_dptr,#addr

        0f8     c,      0f8     c,      039     c,      \ mov_rn,a

        0f8     c,      078     c,      03a     c,      \ mov_rn,#byte

        0f8     c,      0a8     c,      03b     c,      \ mov_rn,byte

        0fe     c,      0f6     c,      03c     c,      \ mov_@ri,a

        0fe     c,      076     c,      03d     c,      \ mov_@ri,#byte

        0fe     c,      0a6     c,      03e     c,      \ mov_@ri,byte

        0ff     c,      0f5     c,      03f     c,      \ mov_byte,a

        0ff     c,      092     c,      040     c,      \ mov_bit,c

        0f8     c,      088     c,      041     c,      \ mov_byte,rn

        0fe     c,      086     c,      042     c,      \ mov_byte,@ri

        0ff     c,      075     c,      043     c,      \ mov_byte,#byte

        0ff     c,      085     c,      044     c,      \ mov_byte,byte

        0ff     c,      093     c,      045     c,      \ movc_a,@a+dptr

        0ff     c,      083     c,      046     c,      \ movc_a,@a+pc

        0fe     c,      0e2     c,      047     c,      \ movx_a,@ri

        0ff     c,      0e0     c,      048     c,      \ movx_a,@dptr

        0fe     c,      0f2     c,      049     c,      \ movx_@ri,a

        0ff     c,      0f0     c,      04a     c,      \ movx_@dptr,a

        0ff     c,      0a4     c,      04b     c,      \ mul_ab

        0ff     c,      000     c,      04c     c,      \ nop

        0f8     c,      048     c,      04d     c,      \ orl_a,rn

        0fe     c,      046     c,      04e     c,      \ orl_a,@ri

        0ff     c,      044     c,      04f     c,      \ orl_a,#byte

        0ff     c,      045     c,      050     c,      \ orl_a,byte

        0ff     c,      0a0     c,      051     c,      \ orl_c,/bit

        0ff     c,      072     c,      052     c,      \ orl_c,bit

        0ff     c,      042     c,      053     c,      \ orl_byte,a

        0ff     c,      043     c,      054     c,      \ orl_byte,#byte

        0ff     c,      0d0     c,      055     c,      \ pop_byte

        0ff     c,      0c0     c,      056     c,      \ push_byte

        0ff     c,      022     c,      057     c,      \ ret

        0ff     c,      032     c,      058     c,      \ reti

        0ff     c,      023     c,      059     c,      \ rla

        0ff     c,      033     c,      05a     c,      \ rlca

        0ff     c,      003     c,      05b     c,      \ rra

        0ff     c,      013     c,      05c     c,      \ rrca

        0ff     c,      0d3     c,      05d     c,      \ setb_c

        0ff     c,      0d2     c,      05e     c,      \ setb_bit

        0ff     c,      080     c,      05f     c,      \ sjmp_pcrel

        0f8     c,      098     c,      060     c,      \ subb_a,rn

        0fe     c,      096     c,      061     c,      \ subb_a,@ri

        0ff     c,      094     c,      062     c,      \ subb_a,#byte

        0ff     c,      095     c,      063     c,      \ subb_a,byte

        0ff     c,      0c4     c,      064     c,      \ swapa

        0f8     c,      0c8     c,      065     c,      \ xch_a,rn

        0fe     c,      0c6     c,      066     c,      \ xch_a,@ri

        0ff     c,      0c5     c,      067     c,      \ xch_a,byte

        0fe     c,      0d6     c,      068     c,      \ xchd_a,@ri

        0f8     c,      068     c,      069     c,      \ xrl_a,rn

        0fe     c,      066     c,      06a     c,      \ xrl_a,@ri

        0ff     c,      064     c,      06b     c,      \ xrl_a,#byte

        0ff     c,      065     c,      06c     c,      \ xrl_a,byte

        0ff     c,      062     c,      06d     c,      \ xrl_byte,a

        0ff     c,      63      c,      06e     c,      \ xrl_byte,#byte

        000     c,      000     c,      06f     c,      \ illop


        decimal


\   To identify an instruction we take the first byte
\   and search the table on a linear way:
\
\   If the byte anded with the mask is equal to the opcode,
\   the instruction has been found, and the token is passed on the
\   stack.
\
\   If the match has not been sucessfull we proceed to the next entry on
\   the table.
\
\   notes: 1) the mask for illegal opcode is zero, and the illop
\              entry is the last one on the table so that the
\              search will end.
\
\          2) the search is slow in Forth, but very fast and easy to
              code in assembler

   code  opcode->token                ( n1 -- n2 )
                                      \ identify the instruction giving
         clear_labels                 \ a token as a result

         pop    cx                    \  unknown opcode
         mov    bx, # opcode-table

 1 $:    mov    dl, 0  [bx]           \  get mask
         mov    al, cl                \  al= unknown opcode
         and    al, dl                \  al = unknown opcode AND mask
         mov    dl, 1  [bx]           \  real opcode
         cmp    al, dl                \  are they equal ?

         je     2 $                   \  if yes then end

         add    bx, # 3               \  else see the
         jmp    1 $                   \       next entry

 2 $:    mov    al, 2  [bx]           \  get the token
         xor    ah, ah                \  clear high byte
         1push
         end-code


\  Forth  opcode->token  definition
\  given for easy porting
\
\  variable table-pointer
\  variable search-value

\  : opcode->token   ( n1 -- n2 )
\
\        search-value !
\        opcode-table table-pointer !
\        begin
\           search-value @          \ get the byte
\           table-pointer @ c@       \ get the mask
\           and
\           table-pointer @ 1+ c@    \ get the opcode
\           <>
\        while
\           table-pointer @          \ if different
\           3 +                      \ advance to the next entry
\           table-pointer !
\        repeat
\
\        table-pointer @ 2 + c@        \ get the token
\
\    ;
\

  : parse-token      ( n --- )        \ sets #bytes, name and operands

       exec:
                acall_addr              \       000

                add_a,rn                \       001

                add_a,@ri               \       002

                add_a,#byte             \       003

                add_a,byte              \       004

                addc_a,rn               \       005

                addc_a,@ri              \       006

                addc_a,#byte            \       007

                addc_a,byte             \       008

                ajmp_addr               \       009

                anl_a,rn                \       00a

                anl_a,@ri               \       00b

                anl_a,#byte             \       00c

                anl_a,byte              \       00d

                anl_c,/bit              \       00e

                anl_c,bit               \       00f

                anl_byte,a              \       010

                anl_byte,#byte          \       011

                cjne_a,#byte,byte       \       012

                cjne_a,byte,byte        \       013

                cjne_rn,#byte,byte      \       014

                cjne_@ri,#byte,byte     \       015

                clr_a                   \       016

                clr_c                   \       017

                clr_bit                 \       018

                cpl_a                   \       019

                cpl_c                   \       01a

                cpl_bit                 \       01b

                daa                     \       01c

                dec_a                   \       01d

                dec_rn                  \       01e

                dec_@ri                 \       01f

                dec_byte                \       020

                divab                   \       021

                djnz_rn,pcrel           \       022

                djnz_byte,pcrel         \       023

                inc_a                   \       024

                inc_rn                  \       025

                inc_@ri                 \       026

                inc_dptr                \       027

                inc_byte                \       028

                jb_bit,pcrel            \       029

                jbc_bit,pcrel           \       02a

                jc_pcrel                \       02b

                jmp_@a+dptr             \       02c

                jnb_bit,pcrel           \       02d

                jnc_pcrel               \       02e

                jnz_pcrel               \       02f

                jz_pcrel                \       030

                lcall_addr              \       031

                ljmp_addr               \       032

                mov_a,rn                \       033

                mov_a,@ri               \       034

                mov_a,#byte             \       035

                mov_a,byte              \       036

                mov_c,bit               \       037

                mov_dptr,#addr          \       038

                mov_rn,a                \       039

                mov_rn,#byte            \       03a

                mov_rn,byte             \       03b

                mov_@ri,a               \       03c

                mov_@ri,#byte           \       03d

                mov_@ri,byte            \       03e

                mov_byte,a              \       03f

                mov_bit,c              \       040

                mov_byte,rn             \       041

                mov_byte,@ri            \       042

                mov_byte,#byte          \       043

                mov_byte,byte           \       044

                movc_a,@a+dptr          \       045

                movc_a,@a+pc            \       046

                movx_a,@ri              \       047

                movx_a,@dptr            \       048

                movx_@ri,a              \       049

                movx_@dptr,a            \       04a

                mul_ab                  \       04b

                nop                     \       04c

                orl_a,rn                \       04d

                orl_a,@ri               \       04e

                orl_a,#byte             \       04f

                orl_a,byte              \       050

                orl_c,/bit              \       051

                orl_c,bit               \       052

                orl_byte,a              \       053

                orl_byte,#byte          \       054

                pop_byte                \       055

                push_byte               \       056

                ret                     \       057

                reti                    \       058

                rla                     \       059

                rlca                    \       05a

                rra                     \       05b

                rrca                    \       05c

                setb_c                  \       05d

                setb_bit                \       05e

                sjmp_pcrel              \       05f

                subb_a,rn               \       060

                subb_a,@ri              \       061

                subb_a,#byte            \       062

                subb_a,byte             \       063

                swapa                   \       064

                xch_a,rn                \       065

                xch_a,@ri               \       066

                xch_a,byte              \       067

                xchd_a,@ri              \       068

                xrl_a,rn                \       069

                xrl_a,@ri               \       06a

                xrl_a,#byte             \       06b

                xrl_a,byte              \       06c

                xrl_byte,a              \       06d

                xrl_byte,#byte          \       06e

                illop                   \       06f

        ;
             \ end of parse-token
   \
   \  In parse-token a word found only in F-PC is used:
   \
   \    exec: ( n -- )
   \                 Executes the n-th word placed ahead on the
   \                 definition.
   \                 The numbers start at zero !! ( the right way )
   \                 So there are trillions of ways of implementing
   \                 parse-word in standard Forth-83.



   headers        \ word used to hide definitions



 :  analyse              ( -- )     \ analyse the opcode pointed by
                                    \  inst-pointer
     inst-pointer @
     xc@
     opcode->token
     parse-token
       ;


 :  dump-instr                 ( -- )   \ dump of the instruction bytes

     base @ >r                            \ save the current base
     hex
           inst-pointer @
            .WORD space  3 spaces             \ prints the address

           #bytes @ 0 do
                         inst-pointer  @     \ prints all the bytes
                         i +
                         xc@   .BYTE space
                loop
      r> base !                              \ restore of the base
           ;


 : dis-one-instr            ( -- )   \ disassembles one instruction

         analyse
         dump-instr
         20 goto-col
         name
         26 goto-col
         operands
                  ;

 : next-instruction        ( -- )   \ move to the next instruction

         inst-pointer  @
         #bytes  @
          +
         inst-pointer  !  ;


 :  dis                    ( addr -- )     \ disassembles one
                                           \ instruction by keypress
                                           \ until user presses space
         cr
         inst-pointer !
         begin
            dis-one-instr
            cr
            next-instruction
            key BL =
         until    ;

 variable dlimit

 :  dis-a                   ( addr1 addr2 -- )    \ disassembles from addr1
                                                  \ up to addr2
         cr
         dlimit ! inst-pointer !
         begin
            dis-one-instr
            cr
            next-instruction
            inst-pointer @ dlimit @ u>
         until    ;


 :  dis-n                   ( addr n -- )     \ disassembles n
                                              \ instruccions
         cr
         swap inst-pointer !
         0 do
            dis-one-instr
            cr
            next-instruction
         loop    ;


 behead      \ used to hide some definitions


