TAQOZ
( 200113 )
{ Martin_H's Inverse Kinematics converted from PropForth to Tachyon

LOAD STATS:
NAMES: 0647 bytes added
CODE:  0790 bytes added

Execution time:
#94 #78 NEWCNT scara_arm park .LAP 104us

}

pub cells 2* 2* ;

0     := SHOULDER
1     := ELBOW
2      := WRIST
3      := COUNT

\ Size of robot bones in consistent units (mm recommended).
long humerus long ulna long humerus_sq long ulna_sq

\ Coordinate of pen tip in Cartesian space.
long x long y

\ Coordinates in arm scara polar space.
long  shoulder_angle long elbow_angle

\ working variables during IK calculations
long s_w long s_w_sqrt

\ Heater's Integer sqrt function
pub isqrtloop
    BEGIN
        DUP 0EXIT
       SWAP OVER + OR
       SWAP - ROT 2DUP SWAP <=
        IF
            SWAP OVER -
            SWAP ROT SWAP OVER +
        ELSE
            ROT SWAP OVER -
        THEN
        2/ SWAP 2/ 2/
    AGAIN
;

\ Integer square root ( i -- s )
pub isqrt
    0 1 #30 SHL isqrtloop DROP NIP
;

\ squares the input ( i -- i*i )
pub square
    DUP *
;


\ word to long sign extension function.
\ n has junk in the top 16 bits, so sign extend
\ ( n -- n )
: w>l
    DUP
    $FFFF
    AND
    SWAP
    $8000 AND
    IF
        $FFFF0000 +
    THEN
;

\ Values for common angles in brad13
$0016     := DEGREE_ANGLE
$0400     := EIGHTH_ANGLE
$0800     := RIGHT_ANGLE
$1000     := STRAIGHT_ANGLE
$1400     := FIVEEIGHTH_ANGLE

\ Create a set of arrays using a workaround for the missing create does>
TABLE table_atan
\ load array with arctangent table iteration from Beau Schwabe's cordic
$20000000 ,            \ atan(1)   = 45    degrees
$12E4051E ,            \ atan(1/2) = 26.56 degrees
$09FB385B ,            \ atan(1/4)
$051111D4 ,            \ atan(1/8)
$028B0D43 ,            \ atan(1/16)
$0145D7E1 ,            \ atan(1/32)
$00A2F61E ,            \ atan(1/64)
$00517C55 ,            \ atan(1/128)
$0028BE53 ,            \ atan(1/256)
$00145F2F ,            \ atan(1/512)
$000A2F98 ,            \ atan(1/1024)
$000517CC ,            \ atan(1/2048)
$00028BE6 ,            \ atan(1/4096)


pub atan() cells table_atan + ;



long servo_center COUNT cells ALLOT
pub servo_center() cells servo_center + ;

long servo_min COUNT cells ALLOT
pub servo_min() cells servo_min + ;

long servo_min COUNT cells ALLOT
pub servo_min() cells servo_min + ;


long servo_max COUNT cells ALLOT
pub servo_max() cells servo_max + ;

long servo_pin COUNT cells ALLOT
pub servo_pin() cells servo_pin + ;

long width_per_radian COUNT cells ALLOT
pub width_per_radian() cells width_per_radian + ;


\ sin function using brads ( brad13 -- sin32 )
\ The sine table contains one quadrant, so we'll need to
\ work around that when generating the index into the table.
\ note use um* with the return value!
pub sin
    DUP \ save a copy of input for later.
    DUP RIGHT_ANGLE AND
    IF
        \ ranges from 800 to 001
        NEGATE $FFF AND
    ELSE
        \ ranges from 000 to 7FF
        $7FF AND
    THEN

    \ convert to word count and index into sin table
    2* $E000 + W@

    \ bit shift to 32 bits for future um*
    $0F SHL

    \ use the sign from the preserved value.
    SWAP
    STRAIGHT_ANGLE AND ?NEGATE
;


\ cos defined in terms of sin by coordinate rotation.
pub cos
    RIGHT_ANGLE +
    sin
;

\ stub to allow compilation
pub acos ;

long atan2_theta

\ 13-bit fixed point four-quadrant arctangent. Given Cartesian vector (x, y),
\ finds the angle subtended by the vector and the positive x-axis.
\ x a signed integer
\ y a signed integer
\ ( x y -- brad13 )
pub atan2
    atan2_theta ~
    OVER 0<
    IF
       NEGATE SWAP \ y = -y
       NEGATE SWAP \ x = -x
       STRAIGHT_ANGLE atan2_theta !
    THEN

    #12 0 DO
        DUP 0<
        IF
            OVER I 2/ OVER +      \ calculate y + (x ~> i)
            ROT ROT                  \ push to bottom of stack
            I 2/ -                \ calculate x - (y ~> i)
            SWAP                  \ update y

            atan2_theta @            \ get the approximation and
            I atan() @ -             \ substract the lookup value.
            atan2_theta !            \ assign it back.
        ELSE
            OVER I 2/ OVER -      \ calculate y - (x ~> i)
            ROT ROT                  \ push to bottom of stack
            I 2/ +                \ calculate x + (y ~> i)
            SWAP                  \ update y

            atan2_theta @    \ get the approximation and
            I atan() @ +     \ add the lookup value.
            atan2_theta !    \ assign it back.
        THEN
    LOOP

    atan2_theta @ #19 SHR
;

\ Computes the pulse width that matches the desired joint angle using
\ the constraints of known angle to pulse values.  Servo center is zero,
\ counter clockwise a quarter circle is Pi/2, while clockwise is -Pi/2.
\  angle - brad value
\  index - servo index
\ ( angle index -- )
pub servo_position
    DUP width_per_radian \ ( angle index width_per_radian )
    SWAP servo_center    \ ( angle width_per_radian center )
    ROT ROT                 \ ( center angle width_per_radian )
    * +                  \ (pulse_width)
\ _servo.writeMicroseconds(pulseWidth)
;

\ Used to initialize arm parameters.
\ humerus - shoulder-to-elbow "bone" in mm
\ ulna - elbow-to-wrist "bone" in mm
pub scara_arm \ (humerus ulna -- )
    DUP humerus !
    square humerus_sq !

    \ pre-calculations
    DUP ulna !
    square ulna_sq !
;

\ Sets the should angle member, computes the servo pulse width for that
\ angle, and sets the associated servo.
\ shoulder_angle - 32-bit floating point radian value
\ ( shoulderAngleRads -- )
pub set_shoulder
    \ Rotate the coordinate system by a right angle so that a straight angle
    \ is a full extension of the joint.
    RIGHT_ANGLE -
    DUP
    shoulder_angle !
    SHOULDER servo_position
;

\ Sets elbow angle data member, computes the servo pulse width for that
\ angle, and sets the associated servo.
\ elbowAngle - floating point radian value
\ ( elbowAngle -- )
pub set_elbow
    \ Rotate the coordinate system by a right angle so that a straight angle
    \ is a full extension of the joint.
    STRAIGHT_ANGLE -
    DUP
    elbow_angle !
    ELBOW servo_position
;

\ Park - Move the arm servos to the parked position.  Called during start and
\ stop. By parking the arm the servos will start from a known position and
\ minimize startup lurch.
\ ( -- )
pub park
    RIGHT_ANGLE set_shoulder
    RIGHT_ANGLE set_elbow
;

\ Clamps n between -32,768 to 32,767 to prevent inverse trig function overflow
\ ( n -- clamped_value )
pub clamp
    -#32768 MAX
    #32767 MIN
;

\ SetPosition : Arm positioning routine utilizing inverse kinematics.  Since
\ the arm is resting on a surface Z can only be positive.  Servo movement
\ constraints prevent y from being negative, But X can be a signed value.
\ Note: This must be called before any of the move routines to initialize
\ arm state!
\  x - the side to side displacement.
\  y - the distance out from the base center.
\ ( x y -- )
pub set_position
    \ Save the Cartesian space coordinates.
    y ! x !

    \ Use Pythagorean theorem to calculate shoulder to wrist distance.
    x @ square
    y @ square
    +
    DUP s_w !
    isqrt s_w_sqrt !

    \ s_w angle to centerline
    x @ y @ atan2 \ a1 is on the stack.

    \ s_w angle to humerus, note the need to prevent overflow!
    \ compute (_humerus_sq - ulna_sq + s_w) / (2 * humerus * s_w_sqrt)
    humerus_sq @ ulna_sq @ - s_w @ +
    2 humerus @ * s_w_sqrt @ *
    /
    clamp acos \ a2 is on the stack

    \ shoulder angle = a1 + a2
    + set_shoulder

    \ elbow angle
    \ compute ( humerus_sq + ulna_sq - s_w ) / ( 2 * humerus * ulna )

    humerus_sq @ ulna_sq @ + s_w @ -
    2 humerus @ * ulna @ *
    /
    clamp acos

    \ Set the elbow
    set_elbow
;

\ Arm positioning routine utilizing inverse kinematics.  Moves the arm from
\ the current Y coordinate to the newY passed in.  It maintains all other
\ position state.
\ newY - the new forward/back displacement
\ ( newY -- )
pub set_y
    x @ SWAP
    set_position
;

\ Arm positioning routine utilizing inverse kinematics.  Moves the arm from
\ the current X coordinate to the newX passed in.  It maintains all other
\ position state.
\ newX - the new height above the table
\ ( newX -- )
pub set_x
    y @
    set_position
;

DECIMAL
\ Arm dimensions( mm )
94 \ shoulder-to-elbow "bone" 3.7"
78 \ elbow-to-wrist "bone" 3.1"
scara_arm
park

END
