IFDEF *SIM65*	AC org FORGET *SIM* }
TAQOZ

pub *SIM65* 	PRINT" A SIMULATOR FOR THE 65C02 - 200217.0000 " ;

--- some add-ons
: RES		|< SWAP CLR ;
: SETB		|< SWAP SET ;

: BIT ( addr bit s/r -- ) IF SETB ELSE RES THEN ;
: C@+		C@ + ;


\ 256 bytes PORTS

--- 65C02 REGISTERS
byte AC
byte PS
byte XR
byte YR
byte SP
word PC


--- 65C02 Status register bits
0 := CARRY
1 := ZFLG
2 := INH
3 := DCM
4 := BRKFLG
6 := OV
7 := NEG

: TEST ( adr bit -- flg )	|< SWAP @ AND 0<> ;

: ZERO?		PS ZFLG TEST ;
: ZERO!		PS ZFLG ROT >B 0= BIT ;
: SIGN?		PS NEG TEST ;
: OVFLOW?	PS OV TEST ;
: CARRY!	PS CARRY ROT BIT ;
: CARRY@	PS CARRY TEST 1& ;
: NEG!		PS NEG ROT BIT ;

: !6502		AC 7 ERASE ;
: AC@		AC C@ ;
: AC!		DUP AC C!
: PS!	( n --- )
	DUP $0100 AND CARRY!
: NZ!	DUP ZERO!  DUP $80 AND NEG!
	DROP
	;
--- !!! OVERFLOW NOT SET !!!


: PC@		PC W@ ;

word _nl ( extra new line when PC jumps )
: NL		PC@ _nl W! ;

: PC+!		PC W+! ;
: PC!		NL PC W! ;

: SP@		SP C@ $0100 + ;

--- modified console output that supports indexed tabs etc
byte xpos
: SIMOUT
		EMIT:
		DUP CONEMIT
		DUP $0A = IF _nl W@
		  ?DUP IF PC@ > IF ." <<<---" ELSE ." --->>>" THEN
		  $0D CONEMIT $0A CONEMIT xpos C~
		  _nl W~ THEN DROP EXIT THEN
		xpos C++
		$0D = IF xpos C~ THEN
		;
: XTAB		BEGIN SPACE DUP xpos C@ $7F AND <= UNTIL DROP ;


byte _list
: LIST?		_list C@ ;
: LISTING	_list C! ;

$7.0000	:= mem6502	--- Use a 64kB area at the end of memory

: @6502		mem6502 + ;
: X@ 		@6502 W@ ;
: XC@		@6502 C@ ;
: XC!		@6502 C! ;
: X!		@6502 W! ;

: .WORD:	.WORD ':' EMIT ;

: RDBYTE	LIST? IF SPACE DUP .WORD: DUP XC@ .BYTE THEN XC@ ;
: RDWORD	LIST? IF SPACE DUP .WORD: DUP X@ .WORD THEN X@ ;

: WRBYTE ( byte addr -- )
		LIST? IF SPACE OVER .BYTE ." ->" DUP .WORD: THEN
		DUP $FC00 > IF 2DROP LIST? IF ."  *ROM* " THEN ELSE XC! THEN
		;

: GETBYTE	PC@ XC@ 1 PC+! ;
: GETWORD	PC@ X@ 2 PC+! ;

: PUSH		SP@ WRBYTE SP C-- ;
: POP		SP C++ SP@ RDBYTE ;
: PUSHPC	PC 1+ C@ PUSH PC C@ PUSH ;

: .PC		PC@ .WORD: SPACE ;

: .DAT		." [" .BYTE ." ] " ;

: .REG
	52 XTAB
	AC@ ." ac" .DAT
	XR C@ ." x" .DAT
	YR C@ ." y" .DAT

	SP C@ ." sp:01" .BYTE SPACE
	8 FOR " czidb-on" I + C@ PS C@ I |< AND IF $20 - THEN EMIT NEXT
	;



--- Addressing modes

: IMM		GETBYTE LIST? IF DUP .BYTE THEN ;
: AB		GETWORD LIST? IF DUP SPACE .WORD THEN ;
: ZPG		GETBYTE ;
: IDX		GETBYTE XR C@+ X@ ;
: IDY		GETBYTE X@ YR C@+ ;
: ZPX		GETBYTE XR C@+ ;
: ZPY		GETBYTE XR C@+ ;
: ABX		GETWORD XR C@+ ;
: ABY		GETWORD YR C@+ ;
: IND		GETBYTE X@ ;

--- instructions

: RDLDA		RDBYTE
: LDA		AC! ;
: LDA.#		IMM LDA ;
: LDA.A		AB RDLDA ;
: LDA.Z		ZPG RDLDA ;
: LDA.X)	IDX RDLDA ;
: LDA.)Y	IDY RDLDA ;
: LDA.Z,X	ZPX RDLDA ;
: LDA.A,X	ABX RDLDA ;
: LDA.A,Y	ABY RDLDA ;
: LDA.)		IND RDLDA ;

: RDADC		RDBYTE
: ADC		CARRY@ + AC@ + LDA ;
: ADC.#		IMM ADC ;
: ADC.A		AB RDADC ;
: ADC.Z		ZPG RDADC ;
: ADC.X)	IDX RDADC ;
: ADC.)Y	IDY RDADC ;
: ADC.Z,X	ZPX RDADC ;
: ADC.A,X	ABX RDADC ;
: ADC.A,Y	ABY RDADC ;
: ADC.)		IND RDADC ;

: RDAND		RDBYTE
: AnD		AC@ AND LDA ;
: AND.#		IMM AnD ;
: AND.A		AB RDAND ;
: AND.Z		ZPG RDAND ;
: AND.X)	IDX RDAND ;
: AND.)Y	IDY RDAND ;
: AND.Z,X	ZPX RDAND ;
: AND.A,X	ABX RDAND ;
: AND.A,Y	ABY RDAND ;
: AND.)		IND RDAND ;

: DOSUB		SWAP 2DUP U< 0= CARRY! - ;

: RDCMP		RDBYTE
: CMP		AC@ DOSUB NZ! ;
: CMP.#		IMM CMP ;
: CMP.A		AB RDCMP ;
: CMP.Z		ZPG RDCMP ;
: CMP.X)	IDX RDCMP ;
: CMP.)Y	IDY RDCMP ;
: CMP.Z,X	ZPX RDCMP ;
: CMP.A,X	ABX RDCMP ;
: CMP.A,Y	ABY RDCMP ;
: CMP.)		IND XC@ CMP ;

: RDEOR		RDBYTE
: EOR		AC@ XOR LDA ;
: EOR.#		IMM EOR ;
: EOR.A		AB RDEOR ;
: EOR.Z		ZPG RDEOR ;
: EOR.X)	IDX RDEOR ;
: EOR.)Y	IDY RDEOR ;
: EOR.Z,X	ZPX RDEOR ;
: EOR.A,X	ABX RDEOR ;
: EOR.A,Y	ABY RDEOR ;
: EOR.)		IND RDEOR ;

: RDORA		RDBYTE
: ORA		AC@ OR LDA ;
: ORA.#		IMM ORA ;
: ORA.A		AB RDORA ;
: ORA.Z		ZPG RDORA ;
: ORA.X)	IDX RDORA ;
: ORA.)Y	IDY RDORA ;
: ORA.Z,X	ZPX RDORA ;
: ORA.A,X	ABX RDORA ;
: ORA.A,Y	ABY RDORA ;
: ORA.)		IND RDORA ;

: RDSBC		RDBYTE
: SBC		CARRY@ 0= +
		AC@ DOSUB LDA ;
: SBC.#		IMM SBC ;
: SBC.A		AB RDSBC ;
: SBC.Z		ZPG RDSBC ;
: SBC.X)	IDX RDSBC ;
: SBC.)Y	IDY RDSBC ;
: SBC.Z,X	ZPX RDSBC ;
: SBC.A,X	ABX RDSBC ;
: SBC.A,Y	ABY RDSBC ;
: SBC.)		IND RDSBC ;

: STA		AC@ SWAP WRBYTE ;
: STA.Z,Y	ZPY STA ;
: STA.A		AB STA ;
: STA.Z		ZPG STA ;
: STA.X)	IDX STA ;
: STA.)Y	IDY STA ;
: STA.Z,X	ZPX STA ;
: STA.A,X	ABX STA ;
: STA.A,Y	ABY STA ;
: STA.)		IND STA ;

: STZ		0 SWAP WRBYTE ;
: STZ.A		AB STZ ;
: STZ.Z		ZPG STZ ;
: STZ.Z,X	ZPX STZ ;
: STZ.A,X	ABX STZ ;

: STX		XR C@ SWAP WRBYTE ;
: STX.A		AB STX ;
: STX.Z		ZPG STX ;
: STX.Z,Y	ZPY STX ;

: STY		YR C@ SWAP WRBYTE ;
: STY.A		AB STY ;
: STY.Z		ZPG STY ;
: STY.Z,X	ZPX STY ;

: RDLDX		RDBYTE
: LDX		DUP XR C! NZ! ;
: LDX.#		IMM LDX ;
: LDX.A		AB RDLDX ;
: LDX.Z		ZPG RDLDX ;
: LDX.A,Y	ABY RDLDX ;
: LDX.Z,Y	ZPY RDLDX ;

: RDLDY		RDBYTE
: LDY		DUP YR C! NZ! ;
: LDY.#		IMM LDY ;
: LDY.A		AB RDLDY ;
: LDY.Z		ZPG RDLDY ;
: LDY.A,X	ABX RDLDY ;
: LDY.Z,X	ZPX RDLDY ;

: RDCPX		RDBYTE
: CPX		XR C@ DOSUB NZ! ;
: CPX.#		IMM CPX ;
: CPX.A		AB RDCPX ;
: CPX.Z		ZPG RDCPX ;

: RDCPY		RDBYTE
: CPY		YR C@ DOSUB NZ! ;
: CPY.#		IMM CPY ;
: CPY.A		AB RDCPY ;
: CPY.Z		ZPG RDCPY ;

: RDBIT
		DUP $80 AND NEG!
		DUP $40 AND PS OV ROT BIT
		AC@ AND ZERO!
;
: BIT.#		IMM RDBIT ;
: BIT.A		AB RDBIT ;
: BIT.Z		ZPG RDBIT ;
: BIT.Z,X	ZPX RDBIT ;
: BIT.A,X	ABX RDBIT ;

: ASL		DUP XC@ 2*
: AS!		DUP ROT WRBYTE PS! ;
: ASL.A		AB ASL ;
: ASL.Z		ZPG ASL ;
: ASL.Z,X	ZPX ASL ;
: ASL.A,X	ABX ASL ;
: ASL.ac	AC@ 2* LDA ;

: LSR		DUP RDBYTE DUP >R 2/ AS! R>
: CARRY:	1& CARRY! ;
: LSR.A		AB LSR ;
: LSR.Z		ZPG LSR ;
: LSR.Z,X	ZPX LSR ;
: LSR.A,X	ABX LSR ;
: LSR.ac	AC@ DUP >R 2/ LDA R> CARRY: ;

: <CARRY	CARRY@ OR ;
: ROL		DUP RDBYTE 2* <CARRY AS! ;
: ROL.A		AB ROL ;
: ROL.Z		ZPG ROL ;
: ROL.Z,X	ZPX ROL ;
: ROL.A,X	ABX ROL ;
: ROL.ac	AC@ 2* <CARRY LDA ;

: >CARRY	CARRY@ 8<< OR  ;
: ROR		DUP RDBYTE >CARRY DUP >R 2/ AS! R> CARRY: ;
: ROR.A		AB ROR ;
: ROR.Z		ZPG ROR ;
: ROR.Z,X	ZPX ROR ;
: ROR.A,X	ABX ROR ;
: ROR.ac	AC@ >CARRY DUP >R 2/ LDA R> CARRY: ;

: TRB		DUP XC@ AC@ AND NZ!
		AC@ $FF XOR OVER RDBYTE AND SWAP WRBYTE ;
: TRB.A		AB TRB ;
: TRB.Z		ZPG TRB ;

: TSB		DUP XC@ AC@ AND NZ!
		AC@ OVER RDBYTE OR SWAP WRBYTE ;
: TSB.A		AB TSB ;
: TSB.Z		ZPG TSB ;


: INCDEC	OVER C+! C@ NZ! ;
: INC.ac	AC 1 INCDEC ;
: INY		YR 1 INCDEC ;
: INX		XR 1 INCDEC ;

: DEC.ac	AC -1 INCDEC ;
: DEY		YR -1 INCDEC ;
: DEX		XR -1 INCDEC ;

: INC		DUP RDBYTE 1+ >B DUP ROT WRBYTE NZ! ;
: INC.A		AB INC ;
: INC.Z		ZPG INC ;
: INC.Z,X	ZPX INC ;
: INC.A,X	ABX INC ;

: DEC		DUP RDBYTE 1- >B DUP ROT WRBYTE NZ! ;
: DEC.A		AB DEC ;
: DEC.Z		ZPG DEC ;
: DEC.Z,X	ZPX DEC ;
: DEC.A,X	ABX DEC ;

: CLV		PS OV RES ;
: CLI		PS INH RES ;
: CLC		PS CARRY RES ;
: CLD		PS DCM RES ;

: SEC		PS CARRY SETB ;
: SED		PS DCM SETB ;
: SEI		PS INH SETB ;

: SKIP		GETBYTE 7 SIGN PC@ + LIST? IF SPACE .WORD ELSE DROP THEN ;
: BRA		GETBYTE 7 SIGN NL PC+! LIST? IF PC@ SPACE .WORD THEN ;
: 0BRA		PS SWAP TEST IF SKIP ELSE BRA THEN ;
: 1BRA		PS SWAP TEST IF BRA ELSE SKIP THEN ;

: BPL		NEG 0BRA ;
: BMI		NEG 1BRA ;
: BVC		OV 0BRA ;
: BVS		OV 1BRA ;
: BCC		CARRY 0BRA ;
: BCS		CARRY 1BRA ;
: BNE		ZFLG 0BRA ;
: BEQ		ZFLG 1BRA ;


: PHA		AC@ PUSH ;
: PHP		PS C@ PUSH ;
: PHX		XR C@ PUSH ;
: PHY		YR C@ PUSH ;

: PLA		POP LDA ;
: PLP		POP PS C! ;
: PLX		POP LDX ;
: PLY		POP LDY ;

: JMP.A		AB PC! ;
: JMP.)		GETWORD RDWORD PC! ;
: JMP.X)	GETWORD XR C@+ RDWORD PC! ;

: JSR
	LIST? IF PC@ X@ SPACE .WORD SPACE THEN
	PC@ X@ 1 PC+! PUSHPC PC!
	;


: RTI		PLP
: POPPC		POP POP B>W PC! ;
: RTS		POPPC NL 1 PC+! ;


: TAX		AC@ LDX ;
: TAY		AC@ LDY ;
: TSX		SP C@ LDX ;
: TXA		XR C@ LDA ;
: TXS		XR C@ DUP SP C! NZ! ;
: TYA		YR C@ LDA ;

: IRQ		$FFFE 0
: INT		PUSHPC PS C@ OR $EF AND PUSH RDWORD PC! ;
: NMI		$FFFA 0 INT ;

byte brkop
---		grab op          save pc and status with brk flag - use irq vector
: BRK		GETBYTE brkop C! $FFFE BRKFLG |< INT ;
 \  PUSHPC PS C@ BRKFLG |< OR PUSH  $FFFE RDWORD PC! ;

: XXX		CON CRLF ." ILLEGAL INSTRUCTION AT " .PC 1 COGSTOP CRLF CONSOLE ;

: OPCODES
---	0	1	2	3	4	5	6	7
( 00 )	BRK	ORA.X)	XXX	XXX	TSB.Z	ORA.Z	ASL.Z	XXX
( 08 )	PHP	ORA.#	ASL.ac	XXX	TSB.A	ORA.A	ASL.A	XXX
( 10 )	BPL	ORA.)Y	ORA.)	XXX	TRB.Z	ORA.Z,X	ASL.Z,X	XXX
( 18 )	CLC	ORA.A,Y	INC.ac	XXX	TRB.A	ORA.A,X	ASL.A,X	XXX
( 20 )	JSR	AND.X)	XXX	XXX	BIT.Z	AND.Z	ROL.Z	XXX
( 28 )	PLP	AND.#	ROL.ac	XXX	BIT.A	AND.A	ROL.A	XXX
( 30 )	BMI	AND.)Y	AND.)	XXX	BIT.Z,X	AND.Z,X	ROL.Z,X	XXX
( 38 )	SEC	AND.A,Y	DEC.ac	XXX	BIT.A,X	AND.A,X	ROL.A,X	XXX

( 40 )	RTI	EOR.X)	XXX	XXX	XXX	EOR.Z	LSR.Z	XXX
( 48 )	PHA	EOR.#	LSR.ac	XXX	JMP.A	EOR.A	LSR.A	XXX
( 50 )	BVC	EOR.)Y	EOR.)	XXX	XXX	EOR.Z,X	LSR.Z,X	XXX
( 58 )	CLI	EOR.A,Y	PHY	XXX	XXX	EOR.A,X	LSR.A,X	XXX
( 60 )	RTS	ADC.X)	XXX	XXX	STZ.Z	ADC.Z	ROR.Z	XXX
( 68 )	PLA	ADC.#	ROR.ac	XXX	JMP.)	ADC.A	ROR.A	XXX
( 70 )	BVS	ADC.)Y	ADC.)	XXX	STZ.Z,X	ADC.Z,X	ROR.Z,X	XXX
( 78 )	SEI	ADC.A,Y	PLY	XXX	JMP.X)	ADC.A,X	ROR.A,X	XXX

( 80 )	BRA	STA.X)	XXX	XXX	STY.Z	STA.Z	STX.Z	XXX
( 88 )	DEY	BIT.#	TXA	XXX	STY.A	STA.A	STX.A	XXX
( 90 )	BCC	STA.)Y	STA.)	XXX	STY.Z,X	STA.Z,X	STX.Z,Y	XXX
( 98 )	TYA	STA.A,Y	TXS	XXX	STZ.A	STA.A,X	STZ.A,X	XXX
( A0 )	LDY.#	LDA.X)	LDX.#	XXX	LDY.Z	LDA.Z	LDX.Z	XXX
( A8 )	TAY	LDA.#	TAX	XXX	LDY.A	LDA.A	LDX.A	XXX
( B0 )	BCS	LDA.)Y	LDA.)	XXX	LDY.Z,X	LDA.Z,X	LDX.Z,Y	XXX
( B8 )	CLV	LDA.A,Y	TSX	XXX	LDY.A,X	LDA.A,X	LDX.A,Y	XXX

( C0 )	CPY.#	CMP.X)	XXX	XXX	CPY.Z	CMP.Z	DEC.Z	XXX
( C8 )	INY	CMP.#	DEX	XXX	CPY.A	CMP.A	DEC.A	XXX
( D0 )	BNE	CMP.)Y	CMP.)	XXX	XXX	CMP.Z,X	DEC.Z,X	XXX
( D8 )	CLD	CMP.A,Y	PHX	XXX	XXX	CMP.A,X	DEC.A,X	XXX
( E0 )	CPX.#	SBC.X)	XXX	XXX	CPX.Z	SBC.Z	INC.Z	XXX
( E8 )	INX	SBC.#	NOP	XXX	CPX.A	SBC.A	INC.A	XXX
( F0 )	BEQ	SBC.)Y	SBC.)	XXX	XXX	SBC.Z,X	INC.Z,X	XXX
( F8 )	SED	SBC.A,Y	PLX	XXX	XXX	SBC.A,X	INC.A,X	XXX
	NOP
;

: DECODE@	2* ' OPCODES + W@ ;

: CYCLE		PC@ XC@ PC W++  DECODE@ JUMP ;
: CYCLES	OFF LISTING FOR CYCLE NEXT ON LISTING ;

: ID?		CFA>NFA C@++ ADO I C@ '.' = IF SPACE ELSE I C@ EMIT THEN LOOP ;

: STEP
	ON LISTING SIMOUT
	CRLF .PC  SPACE PC@ XC@ DECODE@ ID?
	 CYCLE .REG CON
	\ cnt @ IF src @ @6502  cnt @ DUMP THEN
;

: SS		30
: STEPS		ON LISTING FOR STEP NEXT ;
: GO		BEGIN STEP KEY UNTIL ;

ALIAS STEP S


word _pc
: TRAP ( addr/op -- )
	NULL _pc W!
	BEGIN CYCLE _pc 1+ C@ IF PC@ _pc W@ = ELSE PC@ XC@ _pc C@ = THEN
	KEY OR UNTIL
	CON SS
	;


: RESET		!6502 $FFFC X@ PC! ;

pre GETROM
	[C] FOPEN @FILE
	  IF @FILE 0 @6502 64 KB SDRDS ."  Success!" $FF00 @6502 $100 DUMP
	  ELSE ." Fail "
	  THEN
	RESET
	  ;


END
